from righttyper.righttyper_runtime import get_full_type, get_adjusted_full_type
from collections.abc import Iterable
from collections import namedtuple
from typing import Any
import pytest
import importlib


class IterableClass(Iterable):
    def __iter__(self):
        return None


def test_get_full_type():
    assert "bool" == get_full_type(True)
    assert "bool" == get_full_type(False)
    assert "int" == get_full_type(10)
    assert "float" == get_full_type(0.0)
    assert "str" == get_full_type('foo')

    assert "str" == get_full_type(bin(0))
    assert "bool" == get_full_type(bool(0))

    assert "bytearray" == get_full_type(bytearray(b'0000'))
    assert "bytes" == get_full_type(bytes(b'0000'))
    assert "complex" == get_full_type(complex(1, 1))
    assert "list[str]" == get_full_type(dir())

    assert "list[str]" == get_full_type(['a', 'b'])
    assert "list[int]" == get_full_type([0, 1])
    assert "list[tuple[int]]" == get_full_type([(0,), (1,)])

    assert "list[int]" == get_full_type([0, 1][:1])
    assert "int" == get_full_type([0, 1][0])

    #assert "List[int]" == get_full_type([0, 'a'])

    assert "set[str]" == get_full_type({'a', 'b'})
    assert "set[int]" == get_full_type({0, 1})

    # FIXME use Set instead?  specify element type?
    assert "frozenset" == get_full_type(frozenset({'a', 'b'}))
    assert "frozenset" == get_full_type(frozenset({0, 1}))
    assert "frozenset" == get_full_type(frozenset())

    assert "dict[str, str]" == get_full_type({'a': 'b'})

    assert "typing.KeysView[str]" == get_full_type({'a':0, 'b':1}.keys())
    assert "typing.ValuesView[int]" == get_full_type({'a':0, 'b':1}.values())
    assert "typing.ItemsView[str, int]" == get_full_type({'a':0, 'b':1}.items())

    assert "typing.KeysView[typing.Never]" == get_full_type(dict().keys())
    assert "typing.ValuesView[typing.Never]" == get_full_type(dict().values())
    assert "typing.ItemsView[typing.Never, typing.Never]" == get_full_type(dict().items())

    assert "set[str]" == get_full_type({'a', 'b'})
    assert "set[typing.Never]" == get_full_type(set())

    o : Any = range(10)
    assert "range" == get_full_type(o)
    assert 0 == next(iter(o)), "changed state"

    o = iter(range(10))
    assert "typing.Iterator[int]" == get_full_type(o)
    assert 0 == next(o), "changed state"

    o = iter([0,1])
    assert "typing.Iterator[typing.Any]" == get_full_type(o)
    assert 0 == next(o), "changed state"

    o = enumerate([0,1])
    assert "enumerate" == get_full_type(o)
    assert (0, 0) == next(o), "changed state"

    o = filter(lambda x:True, [0,1])
    assert "filter" == get_full_type(o)
    assert 0 == next(o), "changed state"

    o = reversed([0,1])
    assert "typing.Iterator[typing.Any]" == get_full_type(o)
    assert 1 == next(o), "changed state"

    o = zip([0,1], ['a','b'])
    assert "zip" == get_full_type(o)
    assert (0,'a') == next(o), "changed state"

    o = map(lambda x:x, [0,1])
    assert "map" == get_full_type(o)
    assert 0 == next(o), "changed state"

    o = iter({0, 1})
    assert "typing.Iterator[typing.Any]" == get_full_type(o)
    assert 0 == next(o), "changed state"

    o = iter({0:0, 1:1})
    assert "typing.Iterator[typing.Any]" == get_full_type(o)
    assert 0 == next(o), "changed state"

    o = iter({0:0, 1:1}.items())
    assert "typing.Iterator[typing.Any]" == get_full_type(o)
    assert (0, 0) == next(o), "changed state"

    o = iter({0:0, 1:1}.values())
    assert "typing.Iterator[typing.Any]" == get_full_type(o)
    assert 0 == next(o), "changed state"

    o = iter({0:0, 1:1}.keys())
    assert "typing.Iterator[typing.Any]" == get_full_type(o)
    assert 0 == next(o), "changed state"

    o = iter({0:0, 1:1}.items())
    assert "typing.Iterator[typing.Any]" == get_full_type(o)
    assert (0, 0) == next(o), "changed state"

    o = iter({0:0, 1:1}.values())
    assert "typing.Iterator[typing.Any]" == get_full_type(o)
    assert 0 == next(o), "changed state"

    o = (i for i in range(10))
    assert "typing.Generator[typing.Any, typing.Any, typing.Any]" == get_full_type(o)
    assert 0 == next(o), "changed state"

    Point = namedtuple('Point', ['x', 'y'])
    assert f"{__name__}.Point" == get_full_type(Point(1,1))

    assert f"{__name__}.IterableClass" == get_full_type(IterableClass())
    assert "super" == get_full_type(super(IterableClass))

    assert "slice" == get_full_type(slice(0, 5, 1))
    assert "type" == get_full_type(type(str))
    assert "super" == get_full_type(super(str))

    async def async_range(start):
        for i in range(start):
            yield i

    assert "typing.AsyncGenerator[typing.Any, typing.Any]" == get_full_type(async_range(10))
    assert "typing.AsyncGenerator[typing.Any, typing.Any]" == get_full_type(aiter(async_range(10)))


@pytest.mark.filterwarnings("ignore:coroutine .* never awaited")
def test_get_full_type_coro():
    async def coro():
        import asyncio
        await asyncio.sleep(1)

    assert "typing.Coroutine[typing.Any, typing.Any, typing.Any]" == get_full_type(coro())


@pytest.mark.skipif(importlib.util.find_spec('numpy') is None, reason='missing module numpy')
def test_get_full_type_dtype():
    import numpy as np

    assert "numpy.ndarray[typing.Any, numpy.dtypes.Float64DType]" == get_full_type(np.array([], np.float64))


class NonArrayWithDtype:
    def __init__(self):
        self.dtype = 10

def test_non_array_with_dtype():
    # RightTyper used to just check for the presence of a 'dtype' attribute, causing
    # it to generate "mock.MagicMock[Any, mock.MagicMock]" annotations
    assert f"{__name__}.NonArrayWithDtype" == get_full_type(NonArrayWithDtype())


class NamedTupleClass:
    P = namedtuple('P', [])

@pytest.mark.xfail(reason='Not sure how to fix')
def test_get_full_type_namedtuple_in_class():
    # namedtuple's __qualname__ also doesn't contain the enclosing class name...
    assert f"{__name__}.NamedTupleClass.P" == get_full_type(NamedTupleClass.P())


class MyDict(dict):
    def items(self):
        for k, v in super().items():
            yield k, v

def test_get_full_type_dict_with_non_collection_items():
    assert f"{__name__}.MyDict[str, int]" == get_full_type(MyDict({'a': 0}))


class MyList(list):
    pass
class MySet(set):
    pass

def test_get_full_type_custom_collection():
    assert f"{__name__}.MyList[int]" == get_full_type(MyList([0,1]))
    assert f"{__name__}.MySet[int]" == get_full_type(MySet({0,1}))


class Foo:
    pass

def test_adjusted_full_type():
    # these types used to be special cased... ensure they still work
    assert "None" == get_adjusted_full_type(None)
    assert "bool" == get_adjusted_full_type(True)
    assert "float" == get_adjusted_full_type(.0)
    assert "int" == get_adjusted_full_type(0)

    # get_adjusted_full_type's main function is to translate to 'Self'

    class Bar:
        pass

    assert "typing.Self" == get_adjusted_full_type(Foo(), Foo)
    assert f"{__name__}.Foo" == get_adjusted_full_type(Foo())

    assert f"{__name__}.test_adjusted_full_type.<locals>.Bar" == get_adjusted_full_type(Bar())



import sys
from righttyper.righttyper_tool import setup_tool_id, reset_monitoring


def test_reset_monitoring_before_init():
    reset_monitoring()  # should not throw


def test_reset_monitoring_after_init():
    setup_tool_id()
    reset_monitoring()



import textwrap
import libcst as cst
from righttyper.generate_stubs import PyiTransformer


def generate_stub(orig_code: str) -> str:
    m = cst.parse_module(orig_code)
#    print(m)
    m = m.visit(PyiTransformer())
#    print(m)
    return m.code


def test_stubs(tmp_path, monkeypatch):
    code = textwrap.dedent("""\
        import sys

        A = B = 42
        CALC = 1+1
        CALC += 2

        # blah blah blah

        class C:
            '''blah blah blah'''
            class D:
                PI = 314

            def __init__(self: Self, x: int) -> None:  # initializes me
                self.x = x

            def f(self: Self) -> int:
                return self.x

        def f(x: int) -> int:
            return C(x).f()
        """
    )

    output = generate_stub(code)
    assert output == textwrap.dedent("""\
        import sys
        from typing import Any
        A: int
        B: int
        CALC: Any
        class C:
            class D:
                PI: int
            def __init__(self: Self, x: int) -> None: ...
            def f(self: Self) -> int: ...
        def f(x: int) -> int: ...
        """)

def test_stubs_no_any(tmp_path, monkeypatch):
    code = textwrap.dedent("""\
        import sys

        A = 42

        def f(x: int) -> int:
            return C(x).f()
        """
    )

    output = generate_stub(code)
    assert output == textwrap.dedent("""\
        import sys
        A: int
        def f(x: int) -> int: ...
        """)


def test_stubs_assign_tuple(tmp_path, monkeypatch):
    code = textwrap.dedent("""\
        X, Y, Z = 'a', 10, .0
        """
    )

    output = generate_stub(code)
    assert output == textwrap.dedent("""\

        """)


def test_stubs_empty_class(tmp_path, monkeypatch):
    code = textwrap.dedent("""\
        class Foo:
            '''Maybe one day we'll write more'''

        def f(x: int) -> int:
            return 42
        """
    )

    output = generate_stub(code)
    assert output == textwrap.dedent("""\
        class Foo:
            pass
        def f(x: int) -> int: ...
        """)


def test_stubs_conditional(tmp_path, monkeypatch):
    code = textwrap.dedent("""\
        from typing import TYPE_CHECKING
        if TYPE_CHECKING:
            "this should go away"
            import ast

        def f(x: "ast.AST") -> int:
            return 42
        """
    )

    output = generate_stub(code)
    assert output == textwrap.dedent("""\
        from typing import TYPE_CHECKING
        if TYPE_CHECKING:
            import ast
        def f(x: "ast.AST") -> int: ...
        """)


def test_stubs_context_handler(tmp_path, monkeypatch):
    code = textwrap.dedent("""\
        with something():
            "this should go away"
            import ast

        def f(x: "ast.AST") -> int:
            return 42
        """
    )

    output = generate_stub(code)
    assert output == textwrap.dedent("""\
        with something():
            import ast
        def f(x: "ast.AST") -> int: ...
        """)



def test_stubs_try(tmp_path, monkeypatch):
    code = textwrap.dedent("""\
        try:
            from foo import bar
        except ImportError:
            import foobar as bar

        def f(x: bar) -> int:
            return 42
        """
    )

    output = generate_stub(code)
    assert output == textwrap.dedent("""\
        try:
            from foo import bar
        except ImportError:
            import foobar as bar
        def f(x: bar) -> int: ...
        """)


def test_stubs_all_variable(tmp_path, monkeypatch):
    # __all__ is included in many typeshed "pyi"s.
    code = textwrap.dedent("""\
        __all__ = [
            "foo",
            "Bar"
        ]

        def foo() -> int:
            return 42

        class Bar(object):
            def __init__(self, x):
                pass

        def baz() -> float:
            pass
        """
    )

    output = generate_stub(code)
    assert output == textwrap.dedent("""\
        __all__ = [
            "foo",
            "Bar"
        ]
        def foo() -> int: ...
        class Bar(object):
            def __init__(self, x): ...
        def baz() -> float: ...
        """)



from typing import Self
import collections.abc as abc
import libcst as cst


class PyiTransformer(cst.CSTTransformer):
    def __init__(self: Self) -> None:
        self._needs_any = False

    def value2type(self: Self, value: cst.CSTNode) -> str:
        # FIXME not exhaustive; this should come from RightTyper typing
        if isinstance(value, cst.BaseString):
            return str.__name__
        elif isinstance(value, cst.Integer):
            return int.__name__
        elif isinstance(value, cst.Float):
            return float.__name__
        elif isinstance(value, cst.Tuple):
            return tuple.__name__
        elif isinstance(value, cst.BaseList):
            return list.__name__
        elif isinstance(value, cst.BaseDict):
            return dict.__name__
        elif isinstance(value, cst.BaseSet):
            return set.__name__
        self._needs_any = True
        return "Any"

    def handle_body(self: Self, body: abc.Sequence[cst.CSTNode]) -> list[cst.CSTNode]:
        result: list[cst.CSTNode] = []
        for stmt in body:
            if isinstance(stmt, (cst.FunctionDef, cst.ClassDef, cst.If, cst.Try, cst.With)):
                result.append(stmt)
            elif (isinstance(stmt, cst.SimpleStatementLine) and
                  isinstance(stmt.body[0], (cst.Import, cst.ImportFrom))):
                result.append(stmt)
            elif (isinstance(stmt, cst.SimpleStatementLine) and isinstance(stmt.body[0], cst.Assign)):
                if not all(isinstance(target.target, cst.Name) for target in stmt.body[0].targets):
                    # can't handle tuples... do we need to?
                    continue

                if any (isinstance(target.target, cst.Name) and target.target.value == '__all__'
                        for target in stmt.body[0].targets):
                    result.append(stmt)
                    continue

                for target in stmt.body[0].targets:
                    type_ann = self.value2type(stmt.body[0].value)

                    result.append(cst.SimpleStatementLine(body=[
                        cst.AnnAssign(
                            target=target.target,
                            annotation=cst.Annotation(cst.Name(type_ann)),
                            value=None
                        )
                    ]))

        return result

    def leave_FunctionDef(
        self: Self,
        original_node: cst.FunctionDef,
        updated_node: cst.FunctionDef
    ) -> cst.FunctionDef:
        return updated_node.with_changes(
            body=cst.SimpleStatementSuite([cst.Expr(cst.Ellipsis())]),
            leading_lines=[]
        )

    def leave_Comment(    # type: ignore[override]
        self: Self,
        original_node: cst.Comment,
        updated_node: cst.Comment
        ) -> cst.RemovalSentinel:
        return cst.RemoveFromParent()

    def leave_ClassDef(
        self: Self,
        original_node: cst.ClassDef,
        updated_node: cst.ClassDef
    ) -> cst.ClassDef:
        return updated_node.with_changes(
            body=updated_node.body.with_changes(
                body=self.handle_body(updated_node.body.body)
            ),
            leading_lines=[]
        )

    def leave_If(
        self: Self,
        original_node: cst.If,
        updated_node: cst.If
    ) -> cst.If:
        return updated_node.with_changes(
            body=updated_node.body.with_changes(
                body=self.handle_body(updated_node.body.body)
            )
        )

    def leave_With(
        self: Self,
        original_node: cst.With,
        updated_node: cst.With
    ) -> cst.With:
        return updated_node.with_changes(
            body=updated_node.body.with_changes(
                body=self.handle_body(updated_node.body.body)
            )
        )

    def leave_Module(
        self: Self,
        original_node: cst.Module,
        updated_node: cst.Module
    ) -> cst.Module:
        updated_node = updated_node.with_changes(
            body=self.handle_body(updated_node.body)
        )

        if self._needs_any:
            imports = [
                i for i, stmt in enumerate(updated_node.body)
                if (isinstance(stmt, cst.SimpleStatementLine) and
                    isinstance(stmt.body[0], (cst.Import, cst.ImportFrom)))
            ]

            # TODO could check if it's already there
            position = imports[-1]+1 if imports else 0

            updated_node = updated_node.with_changes(
                body=(*updated_node.body[:position],
                      cst.SimpleStatementLine([
                          cst.ImportFrom(
                            module=cst.Name('typing'),
                            names=[cst.ImportAlias(cst.Name('Any'))]
                          ),
                      ]),
                      *updated_node.body[position:]
                )
            )

        return updated_node



import typing
import builtins
import collections.abc as abc
import types
import libcst as cst
import libcst.matchers as cstm
import re

from righttyper.righttyper_types import (
    ArgumentName,
    Filename,
    FuncInfo,
    FuncAnnotation,
    FunctionName,
    Typename,
)


_BUILTIN_TYPES : frozenset[str] = frozenset({
    t for t in (
        "None",
        *(name for name, value in builtins.__dict__.items() if isinstance(value, type))
    )
})

# FIXME this prevents us from missing out on well-known "typing." types,
# but is risky... change to receiving fully qualified names and simplifying
# them in context.
_TYPING_TYPES : frozenset[str] = frozenset({
    t for t in typing.__all__
})

# Regex for a type hint comment
_TYPE_HINT_COMMENT = re.compile(
    r"#\stype:\s*[^\s]+"
)


# Regex for a function and retval type hint comment
_TYPE_HINT_COMMENT_FUNC = re.compile(
    r"#\stype:\s*\([^\)]*\)\s*->\s*[^\s]+"
)


def _dotted_name_to_nodes(name: str) -> cst.Attribute | cst.Name:
    """Creates Attribute/Name to build a module name, dotted or not."""
    parts = name.split(".")
    if len(parts) == 1:
        return cst.Name(parts[0])
    base: cst.Name | cst.Attribute = cst.Name(parts[0])
    for part in parts[1:]:
        base = cst.Attribute(value=base, attr=cst.Name(part))
    return base

def _nodes_to_dotted_name(node: cst.Attribute|cst.Name|cst.BaseExpression) -> str:
    """Extracts a module name from CST Attribute/Name nodes."""
    if isinstance(node, cst.Attribute):
        return f"{_nodes_to_dotted_name(node.value)}.{_nodes_to_dotted_name(node.attr)}"

    assert isinstance(node, cst.Name), f"{node=}"
    return node.value

def _nodes_to_top_level_name(node: cst.Attribute|cst.Name|cst.BaseExpression) -> str:
    """Extracts the top-level name (e.g., 'foo' in 'foo.bar') from CST Attribute/Name nodes."""
    while isinstance(node, cst.Attribute):
        node = node.value

    assert isinstance(node, cst.Name), f"{node=}"
    return node.value

def _nodes_to_all_dotted_names(node: cst.Attribute|cst.Name|cst.BaseExpression) -> list[str]:
    """Extracts the list of all module and parent module names from CST Attribute/Name nodes."""
    if isinstance(node, cst.Attribute):
        names = _nodes_to_all_dotted_names(node.value)
        return [*names, f"{names[-1]}.{node.attr.value}"]

    assert isinstance(node, cst.Name), f"{node=}"
    return [node.value]

def _get_str_attr(obj: object, path: str) -> str|None:
    """Looks for a str-valued attribute along the given dot-separated attribute path."""
    for elem in path.split('.'):
        if obj and isinstance(obj, (list, tuple)):
            obj = obj[0]

        if (obj := getattr(obj, elem, None)) is None:
            break

    return obj if isinstance(obj, str) else None

def _annotation_as_string(annotation: cst.BaseExpression) -> str:
    return cst.Module([cst.SimpleStatementLine([cst.Expr(annotation)])]).code.strip('\n')


class UnifiedTransformer(cst.CSTTransformer):
    def __init__(
        self,
        filename: str,
        type_annotations: dict[FuncInfo, FuncAnnotation],
        override_annotations: bool,
        module_name: str|None,
        module_names: list[str]
    ) -> None:
        self.filename = filename
        self.type_annotations = type_annotations
        self.override_annotations = override_annotations
        self.has_future_annotations = False
        self.module_name = module_name
        self.module_names = sorted(module_names, key=lambda name: -name.count('.'))

    def _module_for(self, name: str) -> tuple[str, str]:
        """Splits a dot name in its module and qualified name parts."""
        # TODO Ideally we'd want to avoid this and just use the type(x).__module__ information
        # we get from the live objects

        for m in self.module_names:
            if name.startswith(m) and (len(name) == len(m) or name[len(m)] == '.'):
                return m, name[len(m)+1:]

        return '', name

    def _is_valid(self, annotation: str) -> bool:
        """Returns whether the annotation can be parsed."""
        # local names such as foo.<locals>.Bar yield this exception
        try:
            cst.parse_expression(annotation)
            return True
        except cst.ParserSyntaxError:
            return False

    def _rename_types(self, annotation: cst.BaseExpression) -> cst.BaseExpression:
        """Renames types in an annotation based on the module name and on any aliases."""

        class Renamer(cst.CSTTransformer):
            def __init__(self, transformer: 'UnifiedTransformer'):
                self.t = transformer

            def visit_Attribute(self, node: cst.Attribute) -> bool:
                return False    # we read the whole name at once, so don't recurse

            def try_replace(self, node: cst.Name|cst.Attribute) -> cst.Name|cst.Attribute:
                name = _nodes_to_dotted_name(node)

                if name in _BUILTIN_TYPES:
                    if name not in self.t.used_names[-1]:
                        return node

                    name = f"builtins.{name}" # somebody overrode a builtin name(!)

                if a := self.t.aliases.get(name):
                    return _dotted_name_to_nodes(a)

                module, rest = self.t._module_for(name)
                if module:
                    if module == self.t.module_name:
                        return _dotted_name_to_nodes(rest)

                    if a := self.t.aliases.get(module):
                        return _dotted_name_to_nodes(f"{a}.{rest}")

                    # does the package name conflict with other definitions?
                    if name.split('.')[0] in self.t.used_names[-1]:
                        alias = "_rt_" + "_".join(module.split("."))
                        self.t.if_checking_aliases[module] = alias
                        return _dotted_name_to_nodes(alias + ("" if rest == "" else f".{rest}"))

                if module == 'builtins':
                    return _dotted_name_to_nodes(name)

                return node

            def leave_Name(self,
                orig_node: cst.Name,
                updated_node: cst.Name
            ) -> cst.Name|cst.Attribute:
                return self.try_replace(updated_node)

            def leave_Attribute(self,
                orig_node: cst.Attribute,
                updated_node: cst.Attribute
            ) -> cst.Name|cst.Attribute:
                return self.try_replace(updated_node)

        return typing.cast(cst.BaseExpression, annotation.visit(Renamer(self)))


    def _unknown_types(self, types: set[str]) -> abc.Iterator[str]:
        """Yields types among those given that are unknown."""
        for t in types:
            if not (
                t.split('.')[0] in self.known_names
                or self._module_for(t)[0] in self.imported_modules
            ):
                yield t


    def visit_Module(self, node: cst.Module) -> bool:
        # Initialize mutable members here, just in case transformer gets reused

        # names used somewhere in a module or class scope
        self.used_names: list[set[str]] = [used_names(node)]

        # currently known global names
        self.known_names: set[str] = set(_BUILTIN_TYPES)

        # global aliases from 'from .. import ..' and 'import .. as ..'
        self.aliases: dict[str, str] = {
            f"typing.{t}": t
            for t in _TYPING_TYPES
        }

        # "import ... as ..." within "if TYPE_CHECKING:".
        # TODO read those in as well so as not to duplicate imports
        self.if_checking_aliases: dict[str, str] = dict()

        # modules imported under their own names
        self.imported_modules: set[str] = set()

        self.unknown_types : set[str] = set()
        self.name_stack : list[str] = []

        self.has_future_annotations = any(
            True for match in cstm.findall(
                node,
                cstm.ImportFrom(
                    module=cstm.Name("__future__"),
                    names=[
                        cstm.ImportAlias(
                            name=cstm.Name("annotations")
                        )
                    ]
                )
            )
        )

        return True


    def visit_If(self, node: cst.If) -> bool:
        if cstm.matches(node, cstm.If(test=cstm.Name("TYPE_CHECKING"))):
            return False    # to ignore imports within TYPE_CHECKING

        return True


    def visit_Import(self, node: cst.Import) -> bool:
        if not self.name_stack: # for now, we only handle global imports
            # node.names could also be cst.ImportStar
            if isinstance(node.names, abc.Sequence):
                for alias in node.names:
                    if alias.asname is not None:
                        self.known_names.add(_nodes_to_top_level_name(alias.asname.name))
                        self.aliases[_nodes_to_dotted_name(alias.name)] = _nodes_to_dotted_name(alias.asname.name)
                    else:
                        self.imported_modules |= set(_nodes_to_all_dotted_names(alias.name))
        return False

    def visit_ImportFrom(self, node: cst.ImportFrom) -> bool:
        if not self.name_stack: # for now, we only handle global imports
            # node.names could also be cst.ImportStar
            if isinstance(node.names, abc.Sequence):
                for alias in node.names:
                    self.known_names.add(
                        _nodes_to_top_level_name(
                            alias.asname.name if alias.asname is not None else alias.name
                        )
                    )

                    source: list[str] = []

                    if node.relative:
                        if self.module_name is None:
                            # TODO warn: cannot handle import: module name unknown
                            return False

                        source = self.module_name.split('.')[:-len(node.relative)]
                        if not source:
                            # TODO warn: invalid import (goes out of package space)
                            return False

                    source += _nodes_to_dotted_name(node.module).split('.') if node.module else []

                    self.aliases[f"{'.'.join(source)}.{_nodes_to_dotted_name(alias.name)}"] = \
                        _nodes_to_dotted_name(
                            alias.asname.name if alias.asname is not None else alias.name
                        )

        return False

    def visit_ClassDef(self, node: cst.ClassDef) -> bool:
        name_source = list_rindex(self.name_stack, '<locals>') # neg. index of last function, or 0 (for globals)
        self.name_stack.append(node.name.value)
        self.used_names.append(self.used_names[name_source] | used_names(node))
        return True

    def leave_ClassDef(self, orig_node: cst.ClassDef, updated_node: cst.ClassDef) -> cst.ClassDef:
        # a class is known once its definition is done
        self.known_names.add(".".join(self.name_stack))
        self.name_stack.pop()
        self.used_names.pop()
        return updated_node

    def visit_FunctionDef(self, node: cst.FunctionDef) -> bool:
        name_source = list_rindex(self.name_stack, '<locals>') # neg. index of last function, or 0 (for globals)
        self.name_stack.extend([node.name.value, "<locals>"])
        self.used_names.append(self.used_names[name_source] | used_names(node))
        return True

    def _process_parameter(self, parameter: cst.Param, ann: FuncAnnotation) -> cst.Param:
        """Processes a parameter, either returning an updated parameter or the original one."""
        if (
            not (parameter.annotation is None or self.override_annotations)
            or (annotation := next(
                (
                    annotation for arg, annotation in ann.args
                    if arg == parameter.name.value
                ), None)
            ) is None
            or not self._is_valid(annotation)
        ):
            return parameter

        annotation_expr: cst.BaseExpression = cst.parse_expression(annotation)
        annotation_expr = self._rename_types(annotation_expr)
        unknown_types = set(self._unknown_types(types_in_annotation(annotation_expr)))
        self.unknown_types |= unknown_types

        if not self.has_future_annotations and (unknown_types - _TYPING_TYPES):
            annotation_expr = cst.SimpleString(f'"{_annotation_as_string(annotation_expr)}"')

        new_par = parameter.with_changes(
            annotation=cst.Annotation(annotation=annotation_expr)
        )

        # remove per-parameter type hint comment for non-last parameter
        if ((comment := _get_str_attr(new_par, "comma.whitespace_after.first_line.comment.value"))
            and _TYPE_HINT_COMMENT.match(comment)):
            new_par = new_par.with_changes(
                comma=new_par.comma.with_changes(   # type: ignore[union-attr]
                    whitespace_after=new_par.comma.whitespace_after.with_changes( # type: ignore[union-attr]
                        first_line=cst.TrailingWhitespace()
                    )
                )
            )

        # remove per-parameter type hint comment for last parameter
        if ((comment := _get_str_attr(new_par, "whitespace_after_param.first_line.comment.value"))
            and _TYPE_HINT_COMMENT.match(comment)):
            new_par = new_par.with_changes(
                whitespace_after_param=new_par.whitespace_after_param.with_changes(
                    first_line=cst.TrailingWhitespace()
                )
            )

        return new_par

    def leave_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> cst.FunctionDef:
        name = ".".join(self.name_stack[:-1])
        self.name_stack.pop()
        self.name_stack.pop()
        self.used_names.pop()
        key = FuncInfo(Filename(self.filename), FunctionName(name))

        if ann := self.type_annotations.get(key):
            for attr_name in ['params', 'kwonly_params', 'posonly_params']:
                if getattr(updated_node.params, attr_name):
                    new_parameters = []
                    for parameter in getattr(updated_node.params, attr_name):
                        new_parameters.append(self._process_parameter(parameter, ann))

                    updated_node = updated_node.with_changes(
                        params=updated_node.params.with_changes(
                            **{attr_name: new_parameters}
                        )
                    )

            for attr_name in ['star_arg', 'star_kwarg']:
                attr = getattr(updated_node.params, attr_name)
                if isinstance(attr, cst.Param):
                    updated_node = updated_node.with_changes(
                        params=updated_node.params.with_changes(
                            **{attr_name: self._process_parameter(attr, ann)}
                        )
                    )

            if ((updated_node.returns is None or self.override_annotations)
                and ann.retval is not None
                and self._is_valid(ann.retval)
            ):
                annotation_expr = cst.parse_expression(ann.retval)
                annotation_expr = self._rename_types(annotation_expr)
                unknown_types = set(self._unknown_types(types_in_annotation(annotation_expr)))
                self.unknown_types |= unknown_types

                if not self.has_future_annotations and (unknown_types - _TYPING_TYPES):
                    annotation_expr = cst.SimpleString(f'"{_annotation_as_string(annotation_expr)}"')

                updated_node = updated_node.with_changes(
                    returns=cst.Annotation(annotation=annotation_expr),
                )

                # remove "(...) -> retval"-style type hint comment
                if ((comment := _get_str_attr(updated_node, "body.body.leading_lines.comment.value"))
                    and _TYPE_HINT_COMMENT_FUNC.match(comment)):
                    updated_node = updated_node.with_changes(
                        body=updated_node.body.with_changes(
                            body=(updated_node.body.body[0].with_changes(leading_lines=[]),
                                  *updated_node.body.body[1:])
                        )
                    )

            # remove single-line type hint comment in the same line as the 'def'
            if ((comment := _get_str_attr(updated_node, "body.header.comment.value"))
                and _TYPE_HINT_COMMENT_FUNC.match(comment)):
                updated_node = updated_node.with_changes(
                    body=updated_node.body.with_changes(
                        header=cst.TrailingWhitespace()))

        return updated_node

    # ConstructImportTransformer logic
    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        # Collect `from __future__` imports and remove them
        future_imports: list[cst.BaseStatement] = []
        new_body: list[cst.BaseStatement] = []
        stmt: cst.BaseStatement

        from_future = cstm.SimpleStatementLine(
            body=[cstm.ImportFrom(
                module=cstm.Name("__future__")
            )]
        )

        for stmt in updated_node.body:
            if cstm.matches(stmt, from_future):
                # Collect future imports to lift later
                future_imports.append(stmt)
            else:
                new_body.append(stmt)

        missing_modules = {
            mod
            for mod in (
                self._module_for(t)[0]
                for t in self.unknown_types - _TYPING_TYPES
                if '.' in t
                and t.split('.')[0] not in self.known_names
            )
            if mod != ''
        }

        def stmt_index(body: list[cst.BaseStatement], pattern: cstm.BaseMatcherNode) -> int|None:
            for i, stmt in enumerate(body):
                if cstm.matches(stmt, pattern):
                    return i

            return None

        def find_beginning(body: list[cst.BaseStatement]) -> int:
            for i, stmt in enumerate(body):
                if not cstm.matches(stmt,
                    cstm.OneOf(
                        cstm.EmptyLine(),
                        cstm.SimpleStatementLine(
                            body=[cstm.Expr(cstm.SimpleString())]
                        )
                    )
                ):
                    break

            return i

        if_type_checking_position = stmt_index(new_body, cstm.If(
                test=cstm.Name('TYPE_CHECKING'),
                body=cstm.IndentedBlock()
            )
        )

        # Add additional type checking imports if needed
        if missing_modules or self.if_checking_aliases:
            existing_body = [*(typing.cast(cst.If, new_body[if_type_checking_position]).body.body
                               if if_type_checking_position is not None
                               else ())]

            # TODO delete modules already imported

            new_stmt: cst.BaseStatement = cst.If(
                test=cst.Name("TYPE_CHECKING"),
                body=cst.IndentedBlock(
                    body=existing_body + [
                        cst.SimpleStatementLine([
                            cst.Import([cst.ImportAlias(_dotted_name_to_nodes(m))])
                        ])
                        for m in sorted(missing_modules)
                    ] + [
                        cst.SimpleStatementLine([
                            cst.Import([cst.ImportAlias(
                                name=_dotted_name_to_nodes(m),
                                asname=cst.AsName(cst.Name(a))
                            )])
                        ])
                        for m, a in sorted(self.if_checking_aliases.items())
                    ]
                )
            )

            if if_type_checking_position is not None:
                new_body[if_type_checking_position] = new_stmt
            else:
                if_type_checking_position = find_beginning(new_body)
                new_body.insert(if_type_checking_position, new_stmt)

            if 'TYPE_CHECKING' not in self.known_names:
                self.unknown_types.add('TYPE_CHECKING')

        # Emit "from typing import ..."
        if (typing_types := (self.unknown_types & _TYPING_TYPES)):
            existing = stmt_index(new_body, cstm.SimpleStatementLine(
                    body=[cstm.ImportFrom(module=cstm.Name("typing"))]
                )
            )

            if existing is not None:
                for alias in cstm.findall(new_body[existing], cstm.ImportAlias()):
                    typing_types.add(_nodes_to_dotted_name(typing.cast(cst.ImportAlias, alias).name))

            new_stmt = cst.SimpleStatementLine(
                body=[
                    cst.ImportFrom(
                        module=cst.Name(value="typing"),
                        names = [
                            cst.ImportAlias(name=_dotted_name_to_nodes(t))
                            # sort 'TYPE_CHECKING' first just because it seems neater that way
                            for t in sorted(typing_types, key=lambda x: (x != 'TYPE_CHECKING', x))
                        ]
                    )
                ]
            )

            if (existing is not None
                and (if_type_checking_position is None
                     or existing < if_type_checking_position)
            ):
                new_body[existing] = new_stmt
            else: 
                position = if_type_checking_position if if_type_checking_position is not None \
                           else find_beginning(new_body)
                new_body.insert(position, new_stmt)


        b = find_beginning(new_body)
        new_body[b:b] = future_imports

        return updated_node.with_changes(body=new_body)


def types_in_annotation(annotation: cst.BaseExpression) -> set[str]:
    """Extracts all type names included in a type annotation."""

    class TypeNameExtractor(cst.CSTVisitor):
        def __init__(self) -> None:
            self.names: set[str] = set()

        def visit_Name(self, node: cst.Name) -> bool:
            self.names.add(node.value)
            return False 

        def visit_Attribute(self, node: cst.Attribute) -> bool:
            self.names.add(_nodes_to_dotted_name(node))
            return False 

    extractor = TypeNameExtractor()
    annotation.visit(extractor)
    return extractor.names


def used_names(node: cst.Module|cst.ClassDef|cst.FunctionDef) -> set[str]:
    """Extracts the names in a module or class."""

    names: set[str] = set()

    class Extractor(cst.CSTVisitor):
        def __init__(self):
            self.in_scope = False

        def visit_Module(self, node: cst.Module) -> bool:
            self.in_scope = True
            return True

        def visit_ClassDef(self, node: cst.ClassDef) -> bool:
            if self.in_scope:
                names.add(node.name.value)
                return False

            self.in_scope = True
            return True

        def visit_FunctionDef(self, node: cst.FunctionDef) -> bool:
            if self.in_scope:
                names.add(node.name.value)
                return False

            self.in_scope = True
            return True

        def visit_Assign(self, node: cst.Assign) -> bool:
            for t in node.targets:
                if isinstance(t.target, cst.Name):
                    names.add(t.target.value)
                elif isinstance(t.target, cst.Tuple):
                    for el in t.target.elements:
                        if isinstance(el.value, cst.Name):
                            names.add(el.value.value)
            return True

        def visit_AnnAssign(self, node: cst.AnnAssign) -> bool:
            if isinstance(node.target, cst.Name):
                names.add(node.target.value)
            return True

        def visit_ImportFrom(self, node: cst.ImportFrom) -> bool:
            # node.names could also be cst.ImportStar
            if isinstance(node.names, abc.Sequence):
                for alias in node.names:
                    if not alias.asname:
                        names.add(_nodes_to_top_level_name(alias.name))
            return True

        def visit_NamedExpr(self, node: cst.NamedExpr) -> bool:
            names.add(_nodes_to_top_level_name(node.target))
            return True

        def visit_AsName(self, node: cst.AsName) -> bool:
            if isinstance(node.name, (cst.Tuple, cst.List)):
                for el in node.name.elements:
                    names.add(_nodes_to_top_level_name(el.value))
            else:
                names.add(_nodes_to_top_level_name(node.name))

            return True

    node.visit(Extractor())
    return names


def list_rindex(l: list, item: object) -> int:
    """Returns either a negative index for the last occurrence of 'item' on the list,
       or 0 if not found."""
    try:
        return -1 - l[::-1].index(item)
    except ValueError:
        return 0



import textwrap
import subprocess
import sys
from pathlib import Path
import pytest
import importlib.util
import re


@pytest.fixture
def tmp_cwd(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    yield tmp_path


@pytest.mark.xfail(reason="value introspection doesn't currently work")
def test_iterable(tmp_cwd):
    t = textwrap.dedent("""\
        def func(iter):
            return enumerate(iter)

        print(list(func(range(10))))
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    
    assert "def func(iter: Iterable[int]) -> Iterable[Tuple[int, int]]" in Path("t.py").read_text()


def test_builtins(tmp_cwd):
    t = textwrap.dedent("""\
        def func(s):
            return range(s.start, s.stop)

        print(list(func(slice(1,3))))

        def func2(t):
            return t.__name__

        print(func2(type(str)))

        def func3(t):
            pass

        func3(super(str))
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()
    
    assert "import slice" not in output
    assert "def func(s: slice) -> range" in output

    assert "import type" not in output
    assert "def func2(t: type) -> str" in output

    assert "import super" not in output
    assert "def func3(t: super) -> None" in output


def test_type_from_generic_alias_annotation(tmp_cwd):
    t = textwrap.dedent("""\
        def f() -> list[int]: ...   # list[int] is a GenericAlias

        def g():
            return f

        g()
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()

    assert "def g() -> Callable[[], list[int]]:" in output


def test_type_from_annotation_none_return(tmp_cwd):
    t = textwrap.dedent("""\
        def f() -> None: ...

        def g():
            return f

        g()
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()

    assert "def g() -> Callable[[], None]:" in output


@pytest.mark.skipif((importlib.util.find_spec('ml_dtypes') is None or
                     importlib.util.find_spec('numpy') is None),
                    reason='missing modules')
def test_numpy_type_name(tmp_cwd):
    t = textwrap.dedent("""\
        import numpy as np
        import ml_dtypes

        def f(t):
            pass

        f(np.dtype(ml_dtypes.bfloat16))
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()

    assert "import bfloat16" not in output
    assert "def f(t: np.dtype[ml_dtypes.bfloat16]) -> None" in output


@pytest.mark.skipif((importlib.util.find_spec('ml_dtypes') is None or
                     importlib.util.find_spec('numpy') is None),
                    reason='missing modules')
def test_numpy_ndarray_dtype_name(tmp_cwd):
    t = textwrap.dedent("""\
        import numpy as np
        import ml_dtypes

        def f(p):
            return str(p)

        bfloat16 = np.dtype(ml_dtypes.bfloat16)
        f(np.array([0], bfloat16))
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()

    assert "import bfloat16\n" not in output
    assert "def f(p: np.ndarray[Any, np.dtype[ml_dtypes.bfloat16]]) -> str" in output


@pytest.mark.skipif((importlib.util.find_spec('ml_dtypes') is None or
                     importlib.util.find_spec('numpy') is None),
                    reason='missing modules')
def test_annotation_with_numpy_dtype_name(tmp_cwd):
    t = textwrap.dedent("""\
        from typing import Any
        import numpy as np
        from ml_dtypes import bfloat16 as bf16

        def f() -> np.ndarray[Any, np.dtype[bf16]]: ...

        def g():
            return f

        g()
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()

    assert "def g() -> Callable[[], np.ndarray[Any, np.dtype[bf16]]]:" in output


@pytest.mark.skipif(importlib.util.find_spec('numpy') is None,
                    reason='missing module numpy')
def test_internal_numpy_type(tmp_cwd):
    t = textwrap.dedent("""\
        import numpy as np
        from numpy.core.overrides import array_function_dispatch

        def sum_dispatcher(arg):
            return (arg,)

        @array_function_dispatch(sum_dispatcher)
        def my_sum(arg):
            return np.sum(arg)

        class MyArray:
            def __init__(self, data):
                self.data = np.array(data)

            def __array_function__(self, func, types, args, kwargs):
                return func(self.data, *args[1:], **kwargs)

        my_sum(MyArray([1, 2, 3, 4]))
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()

    assert (m := re.search(r'__array_function__\(self[^,]*, (func[^,]*),', output))
    assert m.group(1).startswith('func: "numpy.')
    assert m.group(1).endswith('_ArrayFunctionDispatcher"')


def test_call_with_none_default(tmp_cwd):
    t = textwrap.dedent("""\
        def func(n=None):
            return n+1 if n is not None else 0

        func()
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()
    
    assert "def func(n: None=None) -> int" in output


def test_default_arg(tmp_cwd):
    t = textwrap.dedent("""\
        def func(n=None):
            return n+1 if n is not None else 0

        def func2(n=5):
            return n+1

        func(1)
        func2(1.0)
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()
    
    assert "def func(n: int|None=None) -> int" in output

    assert "def func2(n: float|int=5) -> float" in output


def test_function_lookup_for_defaults(tmp_cwd):
    # if it confuses time.time for C.time, an exception is raised, as inspect cannot
    # introspect into time.time
    t = textwrap.dedent("""\
        from time import time

        class C:
            def time(self):
                return 0

        C().time()
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    # FIXME we lack class support
#    output = Path("t.py").read_text()
#    assert "def time(self) -> int" in output


def test_inner_function(tmp_cwd):
    t = textwrap.dedent("""\
        def f(x):
            def g(y):
                return y+1

            return g(x)

        f(1)
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()
    
    assert "def g(y: int) -> int" in output


def test_class_method(tmp_cwd):
    t = textwrap.dedent("""\
        class C:
            def f(self, n):
                return n+1

        def g(x):
            class gC:
                def h(self, x):
                    return x/2

            return gC().h(x)

        C().f(1)
        g(1)
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()
    
    assert "def f(self: Self, n: int) -> int" in output
    assert "\nimport Self" not in output

    assert "def h(self: Self, x: int) -> float" in output


def test_class_method_imported(tmp_cwd):
    Path("m.py").write_text(textwrap.dedent("""\
        class C:
            def f(self, n):
                return n+1

        def g(x):
            class gC:
                def h(self, x):
                    return x/2

            return gC().h(x)
        """
    ))

    Path("t.py").write_text(textwrap.dedent("""\
        import m
        m.C().f(1)
        m.g(1)
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("m.py").read_text()
    
    assert "\nimport Self" not in output

    assert "def f(self: Self, n: int) -> int" in output
    assert "import C" not in output

    assert "def g(x: int) -> float" in output
    assert "def h(self: Self, x: int) -> float" in output
    assert "import gC" not in output


def test_class_name_imported(tmp_cwd):
    Path("m.py").write_text(textwrap.dedent("""\
        class C:
            pass

        def f(x):
            pass

        def g():
            f(C())
        """
    ))

    Path("t.py").write_text(textwrap.dedent("""\
        import m
        m.g()
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("m.py").read_text()
    
    assert "def f(x: C) -> None" in output
    assert "import C" not in output


def test_class_name_in_test(tmp_cwd):
    (tmp_cwd / "tests").mkdir()
    (tmp_cwd / "tests" / "test_foo.py").write_text(textwrap.dedent("""\
        class C:
            pass

        def f(x):
            pass

        def test_foo():
            f(C())
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', '-m', 'pytest', '-s', 'tests'], check=True)
    output = (tmp_cwd / "tests" / "test_foo.py").read_text()
    
    assert "def f(x: C) -> None" in output
    assert "import test_foo" not in output


@pytest.mark.xfail(reason="Doesn't work yet")
def test_local_class_name(tmp_cwd):
    (tmp_cwd / "t.py").write_text(textwrap.dedent("""\
        def f():
            class C:
                pass

            def g(x):
                return 0

            return g(C())

        f()
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = (tmp_cwd / "t.py").read_text()
    
    assert "def g(x: C) -> int" in output


def test_return_private_class(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        def f():
            class fC:
                pass
            return fC()

        def g(x):
            pass

        g(f())
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()
    
    # that local class name is "f.<locals>.fC"; this yields a CST ParserSyntaxError
    assert "import fC" not in output
    assert "def f():" in output # FIXME what is a good way to express the return type?
    assert "def g(x) -> None:" in output # FIXME what is a good way to express the type?


def test_default_inner_function(tmp_cwd):
    t = textwrap.dedent("""\
        def f(x):
            def g(y=None):
                return int(y)+1

            return g(x)

        f(1)
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()
    
    assert "def g(y: int|None=None) -> int" in output


def test_default_class_method(tmp_cwd):
    t = textwrap.dedent("""\
        class C:
            def f(self, n=5):
                return n+1

        def g():
            class gC:
                def h(self, x=1):
                    return x/2

            return gC().h()

        C().f()
        g()
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()
    
    assert "def f(self: Self, n: int=5) -> int" in output
    assert "def h(self: Self, x: int=1) -> float" in output


def test_generator(tmp_cwd):
    t = textwrap.dedent("""\
        def gen():
            yield 10

        def main():
            for _ in gen():
                pass

        def g(f):
            pass

        main()
        g(gen())
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()
    
    # FIXME should be Generator[int] or Iterator[int]
    assert "def gen() -> Generator[int, Any, Any]:" in output
    assert "def g(f: Generator[Any, Any, Any]) -> None" in output


def test_async_generator(tmp_cwd):
    t = textwrap.dedent("""\
        import asyncio

        async def gen():
            yield 10

        async def main():
            async for _ in gen():
                pass

        def g(f):
            pass

        asyncio.run(main())
        g(gen())
        """)

    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("t.py").read_text()
    
    # FIXME should be AsyncGenerator[int] or AsyncIterator[int]
    assert "def gen() -> AsyncGenerator[Any, Any]:" in output
    assert "def g(f: AsyncGenerator[Any, Any]) -> None" in output


def test_generate_stubs(tmp_cwd):
    Path("m.py").write_text(textwrap.dedent("""\
        import sys

        CONST = 42
        CALC = 1+1

        class C:
            PI = 314

            def __init__(self, x):  # initializes me
                self.x = x

            def f(self):
                return self.x

        def f(x):
            return C(x).f()
        """
    ))

    Path("t.py").write_text(textwrap.dedent("""\
        import m
        m.f(1)
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--generate-stubs',
                    '--no-use-multiprocessing', 't.py'], check=True)

    output = Path("m.pyi").read_text()
    # FIXME this assertion is brittle
    assert output == textwrap.dedent("""\
        from typing import Self
        import sys
        from typing import Any
        CONST: int
        CALC: Any
        class C:
            PI: int
            def __init__(self: Self, x: int) -> None: ...
            def f(self: Self) -> int: ...
        def f(x: int) -> int: ...
        """)


def test_type_from_main(tmp_cwd):
    Path("m.py").write_text(textwrap.dedent("""\
        def f(x):
            return str(x)
        """
    ))

    Path("t.py").write_text(textwrap.dedent("""\
        import m

        class C:
            def __str__(self):
                return "hi!"

        m.f(C())
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)
    output = Path("m.py").read_text()
    assert "def f(x: \"t.C\") -> str:" in output

    subprocess.run([sys.executable, '-m', 'mypy', 'm.py', 't.py'], check=True)


def test_coroutine_type(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        def foo():
            async def coro():
                import asyncio
                await asyncio.sleep(1)
            return coro()

        foo()
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert "def foo() -> Coroutine[Any, Any, Any]:" in output


def test_module_type(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        import sys

        def foo(m):
            pass

        foo(sys.modules['__main__'])
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert "def foo(m: \"types.ModuleType\") -> None:" in output
    assert "import types" in output


def test_function_type(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        def foo(x: int) -> float:
            return x/2

        def bar(f, g, x):
            return f(x) + g(C(), x)

        def baz(h, x):
            return h(x)

        class C:
            def foo2(self: "C", x: int) -> float:
                return x*.5

        bar(foo, C.foo2, 1)
        baz(C().foo2, 1)
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert 'def bar(f: Callable[[int], float], g: "Callable[[C, int], float]", x: int) -> float:' in output
    assert 'def baz(h: Callable[[int], float], x: int) -> float:' in output # bound method


def test_function_type_future_annotations(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        from __future__ import annotations

        def foo(x: int) -> float:
            return x/2

        def bar(f, g, x):
            return f(x) + g(C(), x)

        def baz(h, x):
            return h(x)

        class C:
            def foo2(self: "C", x: int) -> int:
                return x//2

        bar(foo, C.foo2, 1)
        baz(C().foo2, 1)
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert "def bar(f: Callable[[int], float], g: Callable[[C, int], int], x: int) -> float:" in output
    assert 'def baz(h: Callable[[int], int], x: int) -> int:' in output # bound method


def test_function_type_in_annotation(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        from types import FunctionType

        def foo(x: int) -> float:
            return x/2

        def bar(g: FunctionType, x):
            return g(x)

        def baz(f, g, x):
            return bar(g, x)

        baz(bar, foo, 1)
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert 'def bar(g: FunctionType, x: int) -> float:' in output
    assert 'def baz(f: Callable[[FunctionType, Any], Any], g: Callable[[int], float], x: int) -> float:' in output


@pytest.mark.xfail(reason="our function types are all annotation-based so far")
def test_discovered_function_type(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        def foo(x):
            return x/2

        def bar(f, x):
            return f(x)

        bar(foo, 1)
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--no-use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert "def foo(x: int) -> float:" in output
    assert "def bar(f: Callable[[int], float], x: int) -> float:" in output


def test_module_list_not_lost_with_multiprocessing(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        def foo(t):
            pass

        from xml.dom.minidom import Element as E
        foo(E('foo'))
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert 'def foo(t: "xml.dom.minidom.Element") -> None:' in output

    assert 'import xml.dom.minidom\n' in output


def test_posonly_and_kwonly(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        def foo(x, /, *, y):
            pass

        foo(10, y=.1)
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert 'def foo(x: int, /, *, y: float) -> None:' in output


def test_varargs(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        def foo(x, *args):
            pass

        foo(True, 10, 's', 0.5)
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert 'def foo(x: bool, *args: float|int|str) -> None:' in output


def test_kwargs(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        def foo(x, **kwargs):
            pass

        foo(True, a=10, b='s', c=0.5)
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert 'def foo(x: bool, **kwargs: float|int|str) -> None:' in output


def test_none_arg(tmp_cwd):
    Path("t.py").write_text(textwrap.dedent("""\
        def foo(x):
            pass

        foo(None)
        """
    ))

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files',
                    '--use-multiprocessing', 't.py'], check=True)

    output = Path("t.py").read_text()
    assert 'def foo(x: None) -> None:' in output



import ast
import os
import platform
import runpy
import site
import subprocess
import sys
import sysconfig
from functools import lru_cache
from importlib.abc import Loader, MetaPathFinder
from typing import Any


@lru_cache()
def get_homebrew_cellar_path() -> str:
    """Returns the Homebrew Cellar path by running `brew --cellar`."""
    try:
        result = subprocess.run(
            ["brew", "--cellar"], capture_output=True, text=True, check=True
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError:
        return ""


@lru_cache()
def get_venv_site_packages() -> str:
    """
    Returns the site-packages path for a virtual environment, based on sys.prefix.
    This function supports both Linux/macOS and Windows.
    """
    # Determine the correct location of site-packages depending on the platform
    if platform.system() == "Windows":
        return os.path.join(sys.prefix, "Lib", "site-packages")
    else:
        return os.path.join(
            sys.prefix, "lib", f"python{sys.version[:3]}", "site-packages"
        )


@lru_cache()
def get_user_site_packages() -> str:
    """Returns the path to the user-specific site-packages directory."""
    return site.getusersitepackages()


@lru_cache()
def is_system_installed_package_file(file_path: str) -> bool:
    """
    Checks whether a given file belongs to a system-installed package, including
    Python's standard library, installed packages in site-packages, platform-specific
    directories such as lib-dynload, Homebrew paths, and virtual environments.

    Args:
        file_path (str): The absolute path to the file to check.

    Returns:
        bool: True if the file belongs to a system-installed package, False otherwise.
    """
    # Normalize the input file path
    file_path = os.path.normcase(os.path.abspath(file_path))

    # Get standard directories where system packages are installed
    package_dirs = [
        sysconfig.get_path("purelib"),  # Site-packages
        sysconfig.get_path("platlib"),  # Platform-specific site-packages
        sysconfig.get_path("stdlib"),  # Standard library
        sysconfig.get_path("platstdlib"),  # Platform-specific standard library
    ]

    # Get the dynamic library directory for platform-specific modules (like _contextvars.so)
    dynload_dir = os.path.join(sysconfig.get_path("platstdlib"), "lib-dynload")
    package_dirs.append(dynload_dir)

    # Also add sys.prefix as a fallback for virtual environments or custom installs
    package_dirs.append(sys.prefix)

    # Detect if we are running on macOS
    if platform.system() == "Darwin":
        # Check if Python is installed via Homebrew by checking if 'brew' is available and valid
        brew_cellar_path = get_homebrew_cellar_path()
        if brew_cellar_path:
            package_dirs.append(brew_cellar_path)

    # Detect if we're in a virtual environment
    if sys.prefix != sys.base_prefix:  # Virtual environment detection
        venv_site_packages = get_venv_site_packages()
        package_dirs.append(venv_site_packages)

    # Add the user-specific site-packages (locally installed packages)
    user_site_packages = get_user_site_packages()
    package_dirs.append(user_site_packages)

    # Normalize and clean the paths for comparison
    package_dirs = [
        os.path.normcase(os.path.abspath(path))
        for path in package_dirs
        if path
    ]

    # Check if the file resides within any of the identified system directories
    for package_dir in package_dirs:
        if file_path.startswith(package_dir):
            return True

    return False


class DictTransformer(ast.NodeTransformer):
    def visit_Module(self, node):
        """Insert import at the top of the module only if necessary."""
        has_random_dict_import = any(
            isinstance(n, ast.ImportFrom)
            and n.module == "righttyper.random_dict"
            for n in node.body
        )

        if not has_random_dict_import:
            import_node = ast.ImportFrom(
                module="righttyper.random_dict",
                names=[ast.alias(name="RandomDict", asname=None)],
                level=0,
            )

            # Find the last future import
            last_future_import_idx = -1
            for idx, n in enumerate(node.body):
                if isinstance(n, ast.ImportFrom) and n.module == "__future__":
                    last_future_import_idx = idx

            # Insert the import statement right after the last __future__ import
            insert_position = last_future_import_idx + 1
            node.body.insert(insert_position, import_node)

        # Continue with other transformations
        self.generic_visit(node)
        return node

    def visit_Dict(self, node):
        # Transform dictionary literals to RandomDict()
        self.generic_visit(node)
        if not node.keys and not node.values:
            # Empty dict literal {}
            return ast.Call(
                func=ast.Name(id="RandomDict", ctx=ast.Load()),
                args=[],
                keywords=[],
            )
        else:
            # Non-empty dict literal {k1: v1, k2: v2, ...}
            dict_call = ast.Dict(keys=node.keys, values=node.values)
            return ast.Call(
                func=ast.Name(id="RandomDict", ctx=ast.Load()),
                args=[dict_call],
                keywords=[],
            )

    def visit_DictComp(self, node):
        # Transform dictionary comprehensions to RandomDict()
        self.generic_visit(node)
        return ast.Call(
            func=ast.Name(id="RandomDict", ctx=ast.Load()),
            args=[node],
            keywords=[],
        )

    def visit_Call(self, node):
        # Replace calls to dict() with RandomDict()
        self.generic_visit(node)
        if isinstance(node.func, ast.Name) and node.func.id == "dict":
            node.func.id = "RandomDict"
        return node


class TransformingLoader(Loader):
    def __init__(self, loader):
        self.loader = loader

    def create_module(self, spec):
        return self.loader.create_module(spec)

    def exec_module(self, module):
        # Apply transformation only if the module is in the user code path
        source = self.loader.get_source(module.__name__)
        if source is None:
            return
        tree = ast.parse(source)
        transformer = DictTransformer()
        tree = transformer.visit(tree)
        ast.fix_missing_locations(tree)
        assert module.__file__ is not None
        code = compile(tree, filename=module.__file__, mode="exec")
        exec(code, module.__dict__)

    def get_code(self, fullname):
        return self.loader.get_code(fullname)


class TransformingFinder(MetaPathFinder):
    def __init__(self) -> None:
        self._processed_modules: set[str] = set()

    def find_spec(self, fullname, path, target=None):
        if fullname in self._processed_modules:
            return None
        self._processed_modules.add(fullname)
        for finder in sys.meta_path:
            if finder is self:
                continue
            spec = finder.find_spec(fullname, path, target)
            if (
                spec
                and spec.origin
                and spec.origin.endswith(os.extsep + "py")
                and not is_system_installed_package_file(spec.origin)
            ):
                spec.loader = TransformingLoader(spec.loader)
                return spec
        return None


def transform_and_run_script(script_path):
    """Transform and run a Python script using AST transformation."""
    with open(script_path, "r") as file:
        source = file.read()

    # Only apply transformation if the script is within the user code path
    if not is_system_installed_package_file(os.path.abspath(script_path)):
        tree = ast.parse(source)
        transformer = DictTransformer()
        tree = transformer.visit(tree)
        ast.fix_missing_locations(tree)
        code = compile(tree, filename=script_path, mode="exec")
        namespace: dict[str, Any] = {}
        exec(code, namespace)
    else:
        # Fallback to default runpy if not in user path
        runpy.run_path(script_path, run_name="__main__")


def replace_dicts():
    """Replace dict with RandomDict and apply transformation to modules and scripts."""
    sys.meta_path.insert(0, TransformingFinder())
    # Monkey-patch runpy.run_path to apply AST transformation to scripts
    original_run_path = runpy.run_path

    def custom_run_path(script_path, init_globals=None, run_name=None):
        if not is_system_installed_package_file(os.path.abspath(script_path)):
            transform_and_run_script(script_path)
        else:
            return original_run_path(
                script_path, init_globals=init_globals, run_name=run_name
            )

    runpy.run_path = custom_run_path



import logging
import os
import pathlib
from collections import defaultdict
from typing import Any

import libcst as cst

from righttyper.generate_stubs import PyiTransformer
from righttyper.righttyper_types import (
    ArgInfo,
    ArgumentName,
    ArgumentType,
    Filename,
    FuncInfo,
    FuncAnnotation,
    FunctionName,
    Typename,
    TypenameSet,
)
from righttyper.righttyper_utils import (
    debug_print,
    make_type_signature,
    skip_this_file,
    union_typeset_str,
)
from righttyper.unified_transformer import UnifiedTransformer
from righttyper.righttyper_runtime import source_to_module_fqn

logger = logging.getLogger("righttyper")


def correct_indentation_issues(file_contents: str) -> str:
    """Return a string corresponding to the file contents, but with indentation issues fixed if needed."""
    original_lines = file_contents.splitlines(
        keepends=True
    )  # Preserve line endings

    indent_stack: list[int] = []
    corrected_lines = []

    for line_number, line in enumerate(original_lines, start=1):
        stripped_line = line.lstrip()

        if not stripped_line or stripped_line.startswith("#"):
            # Preserve empty lines and comments
            corrected_lines.append(line)
            continue

        leading_whitespace = line[: len(line) - len(stripped_line)]

        # Convert tabs to spaces (assuming 4 spaces per tab)
        corrected_leading_whitespace = leading_whitespace.replace(
            "\t", " " * 4
        )

        # Check for mixed tabs and spaces and correct them
        if " " in leading_whitespace and "\t" in leading_whitespace:
            # print(f"Line {line_number}: Mixed spaces and tabs detected. Correcting to spaces.")
            leading_whitespace = corrected_leading_whitespace

        indent_level = len(corrected_leading_whitespace)

        # Check if the indentation level matches the current stack
        if indent_stack and indent_level < indent_stack[-1]:
            # If a dedent is found, adjust the stack accordingly
            while indent_stack and indent_level < indent_stack[-1]:
                indent_stack.pop()

        # Push the new indentation level if a new block starts
        if stripped_line.endswith(":"):  # Detect start of a block
            indent_stack.append(indent_level)

        # Reconstruct the corrected line and append to corrected lines list
        corrected_line = corrected_leading_whitespace + stripped_line
        corrected_lines.append(corrected_line)

    # Join corrected lines into a single string
    corrected_content = "".join(corrected_lines)

    # Return the corrected content if changes were made, otherwise return original content
    if corrected_content != file_contents:
        return corrected_content
    else:
        return file_contents


def process_file(
    filename: Filename,
    output_files: bool,
    generate_stubs: bool,
    type_annotations: dict[FuncInfo, FuncAnnotation],
    overwrite: bool,
    module_names: list[str],
    ignore_annotations: bool = False,
    srcdir: str = "",
) -> None:
    debug_print(f"process_file: {filename}")
    try:
        with open(filename, "r") as file:
            source = file.read()
    except FileNotFoundError:
        return

    if output_files and overwrite:
        with open(filename + ".bak", "w") as file:
            file.write(source)


    try:
        cst_tree = cst.parse_module(source)
    except cst._exceptions.ParserSyntaxError:  # type: ignore
        try:
            # Initial parse failed; fix any indentation issues and try again
            source = correct_indentation_issues(source)
            cst_tree = cst.parse_module(source)
        except cst._exceptions.ParserSyntaxError:  # type: ignore
            print(f"Failed to parse source for {filename}.")
            return

    transformer = UnifiedTransformer(
        filename, type_annotations, ignore_annotations,
        module_name=source_to_module_fqn(pathlib.Path(filename)),
        module_names=module_names
    )

    try:
        transformed = cst_tree.visit(transformer)
    except TypeError as e:
        # This happens when "Mock" is passed around.
        # Print a warning and bail.
        print(f"Failed to transform {filename}. ({e})")
        return

    if output_files:
        with open(
            filename + ("" if overwrite else ".typed"),
            "w",
        ) as file:
            file.write(transformed.code)


    if generate_stubs:
        stub_file = pathlib.Path(filename).with_suffix(".pyi")

        stubs = transformed.visit(PyiTransformer())

        if stub_file.exists():
            stub_file.with_suffix(stub_file.suffix + ".bak").write_text(stub_file.read_text())

        stub_file.write_text(stubs.code)


# Convert the collected data into the expected format for type_annotations
def collect_data(
    file_name: str,
    visited_funcs: set[FuncInfo],
    visited_funcs_arguments: dict[FuncInfo, list[ArgInfo]],
    visited_funcs_retval: dict[FuncInfo, TypenameSet],
    namespace: dict[str, Any] = globals(),
) -> dict[FuncInfo, FuncAnnotation]:
    type_annotations: dict[FuncInfo, FuncAnnotation] = {}
    for t in visited_funcs:
        args = visited_funcs_arguments[t]
        arg_annotations = [
            (
                ArgumentName(arginfo.arg_name),
                union_typeset_str(
                    file_name,
                    arginfo.type_name_set,
                    namespace,
                ),
            )
            for arginfo in args
        ]
        if t in visited_funcs_retval:
            retval = union_typeset_str(
                file_name,
                visited_funcs_retval[t],
                namespace,
            )
        else:
            retval = Typename("None")
        type_annotations[t] = FuncAnnotation(
            arg_annotations,
            retval,
        )
        # print(f"{type_annotations[t]} {t}")
    return type_annotations



import random

__version__ = "0.2.2"

from typing import Any


class RandomDict(dict):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._update_internal_vectors()

    def __new__(cls, *args, **kwargs) -> "RandomDict":
        instance = super().__new__(cls)
        instance._keys = dict()
        instance._random_vector = []
        return instance

    def _update_internal_vectors(self) -> None:
        """Helper method to update _random_vector and _keys."""
        self._random_vector = list(self.keys())  # Rebuild the list of keys
        self._keys = {key: idx for idx, key in enumerate(self._random_vector)}

    def update(self, *args, **kwargs) -> None:
        """Update the dictionary and ensure _random_vector is in sync."""
        super().update(*args, **kwargs)
        self._update_internal_vectors()

    def setdefault(self, key, default=None) -> Any:
        """Override setdefault to ensure consistency of _random_vector."""
        if key not in self:
            self[key] = default
        return self[key]

    def copy(self) -> "RandomDict":
        """Return a shallow copy of the RandomDict"""
        new_rd = RandomDict(super().copy())
        new_rd._keys = self._keys.copy()
        new_rd._random_vector = self._random_vector[:]
        return new_rd

    @classmethod
    def fromkeys(cls, keys, value=None) -> "RandomDict":
        """Create a RandomDict from an iterable of keys, all mapped to the same value."""
        rd = cls()
        for key in keys:
            rd[key] = value
        rd._update_internal_vectors()  # Make sure _random_vector is populated
        return rd

    def __setitem__(self, key, value) -> None:
        """Insert or update a key-value pair"""
        super().__setitem__(key, value)
        i = self._keys.get(key, -1)

        if i == -1:
            # Add new key
            self._random_vector.append(key)
            self._keys[key] = len(self._random_vector) - 1

    def __delitem__(self, key) -> None:
        """Delete item by swapping with the last element in the random vector"""
        if key not in self._keys:
            raise KeyError(key)

        # Get the index of the item to delete
        i = self._keys[key]

        # Remove the last item from the random vector
        move_key = self._random_vector.pop()

        # Only swap if we are not deleting the last item
        if len(self._random_vector) > i:
            # Move the last item into the location of the deleted item
            self._random_vector[i] = move_key
            self._keys[move_key] = i

        del self._keys[key]
        super().__delitem__(key)

    def random_key(self) -> Any:
        """Return a random key from this dictionary in O(1) time"""
        if len(self._random_vector) == 0:
            print(
                f"Debug: _random_vector is empty. Current dict: {self}"
            )  # Debug print statement
            raise KeyError("RandomDict is empty")
        return random.choice(self._random_vector)

    def random_value(self) -> Any:
        """Return a random value from this dictionary in O(1) time"""
        return self[self.random_key()]

    def random_item(self) -> tuple[Any, Any]:
        """Return a random key-value pair from this dictionary in O(1) time"""
        k = self.random_key()
        return k, self[k]


def replace_dicts():
    # Replace dict with RandomDict
    import builtins

    builtins.dict = RandomDict  # type: ignore

    # Replace defaultdict with RandomDict

    # stash the original import for use in a custom importer
    _original_import = builtins.__import__

    def _custom_import(name, globals=None, locals=None, fromlist=(), level=0):
        """Intercept imports of defaultdict to route to RandomDict"""
        module = _original_import(name, globals, locals, fromlist, level)
        if name == "collections" or (fromlist and "defaultdict" in fromlist):
            module.__dict__["defaultdict"] = RandomDict
        return module

    # Monkey-patch __import__
    builtins.__import__ = _custom_import



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, NewType, TypeVar

T = TypeVar("T")

ArgumentName = NewType("ArgumentName", str)


class ArgumentType(Enum):
    positional = "PositionalArgument"
    vararg = "VariableArgument"
    kwarg = "KeywordArgument"


Filename = NewType("Filename", str)
FunctionName = NewType("FunctionName", str)


@dataclass(eq=True, frozen=True)
class FuncInfo:
    file_name: Filename
    func_name: FunctionName

@dataclass(eq=True, frozen=True)
class FuncAnnotation:
    args: list[tuple[ArgumentName, Typename]]
    retval: Typename|None


Typename = NewType("Typename", str)


@dataclass
class TypenameFrequency:
    typename: Typename
    counter: int

    def __hash__(self: TypenameFrequency) -> int:
        return hash(self.typename)

    def __eq__(self: TypenameFrequency, other: Any) -> bool:
        if isinstance(other, TypenameFrequency):
            return self.typename == other.typename
        return False


TypenameSet = NewType("TypenameSet", set[TypenameFrequency])


@dataclass
class ArgInfo:
    arg_name: ArgumentName
    type_name_set: TypenameSet



import inspect
import os
import random
import sys
import collections.abc as abc
from functools import cache
from itertools import islice
from types import CodeType, FrameType, NoneType, FunctionType, MethodType, GenericAlias
from typing import Any
from pathlib import Path

from righttyper.random_dict import RandomDict
from righttyper.righttyper_types import (
    ArgInfo,
    ArgumentName,
    ArgumentType,
    FuncInfo,
    T,
    Typename,
    TypenameFrequency,
    TypenameSet,
)
from righttyper.righttyper_utils import skip_this_file


def sample_from_collection(value: abc.Collection[T]|abc.Iterator[T], depth = 0) -> T:
    """Samples from a collection, or from an interator/generator whose state we don't mind changing."""
    MAX_ELEMENTS = 10   # to keep this O(1)

    if isinstance(value, abc.Collection):
        n = random.randint(0, min(MAX_ELEMENTS, len(value) - 1))
        return next(islice(value, n, n + 1))

    n = random.randint(1, MAX_ELEMENTS)
    return list(islice(value, n))[-1]



@cache
def should_skip_function(
    code: CodeType,
    script_dir: str,
    include_all: bool,
    include_files_regex: str,
) -> bool:
    if (
        code.co_name.startswith("<")
        or skip_this_file(
            code.co_filename,
            script_dir,
            include_all,
            include_files_regex,
        )
        or "righttyper" + os.sep in code.co_filename
    ):
        return True
    if not (code.co_flags & 0x2):
        import dis

        assert dis.COMPILER_FLAG_NAMES[0x2] == "NEWLOCALS"
        return True
    return False


def get_class_type_from_stack(
    max_depth: int = 5,
) -> type|None:
    # Initialize the current frame
    current_frame = inspect.currentframe()
    try:
        # Move up in the stack frame by frame
        depth = 0
        while current_frame and depth < max_depth:
            # Check if 'self' is in the local variables of the frame
            if "self" in current_frame.f_locals:
                instance = current_frame.f_locals["self"]
                return instance.__class__
            # Move one level up in the stack
            current_frame = current_frame.f_back
            depth += 1
    finally:
        # Delete reference to the current frame to avoid reference cycles
        del current_frame
    return None


def type_from_annotations(func: FunctionType | MethodType) -> str:
    # Get the signature of the function
    signature = inspect.signature(func)
    #print(f"{func=} {signature=}")

    # Extract argument types, default to Any if no annotation provided
    arg_types = [
        (param.annotation if param.annotation is not param.empty else Any)
        for name, param in signature.parameters.items()
    ]

    # Extract return type, default to Any if no annotation provided
    return_type = signature.return_annotation
    if return_type is inspect.Signature.empty:
        return_type = Any

    def format_arg(arg) -> str:
        if isinstance(arg, str):
            # If types are quoted while using "from __future__ import annotations",
            # strings may appear double quoted
            if len(arg) >= 2 and arg[0] == arg[-1] and arg[0] in ["'",'"']:
                arg = arg[1:-1]

            return arg

        if isinstance(arg, GenericAlias):
            return str(arg)

        if arg is None:
            return "None"

        return get_type_name(arg)

    # Format the result
    arg_types_str = ", ".join([format_arg(arg) for arg in arg_types])
    return_type_str = format_arg(return_type)

    # Construct the Callable type string
    return f"typing.Callable[[{arg_types_str}], {return_type_str}]"


def find_caller_frame() -> FrameType|None:
    """Attempts to find the stack frame which from which we were called. A bit brittle!"""
    from pathlib import Path

    pkg_path = Path(find_caller_frame.__code__.co_filename).parent

    frame: FrameType|None = sys._getframe(1)

    while (frame is not None and frame.f_code is not None and
           pkg_path in Path(frame.f_code.co_filename).parents):
        frame = frame.f_back

    return frame


@cache
def from_types_import(t: type) -> str | None:
    # TODO we could also simply reverse types.__dict__ ...
    import types

    for k in types.__all__:
        if (v := types.__dict__.get(k)) and t is v:
            return k

    return None


@cache
def in_builtins_import(t: type) -> bool:
    import builtins

    if bt := builtins.__dict__.get(t.__name__):
        return t == bt

    return False


@cache
def lookup_type_module(t: type) -> str:
    parts = t.__qualname__.split('.')

    def is_defined_in_module(namespace: dict, index: int=0) -> bool:
        if index<len(parts) and (obj := namespace.get(parts[index])):
            if obj is t:
                return True

            if isinstance(obj, dict):
                return is_defined_in_module(obj, index+1)

        return False

    if (m := sys.modules.get(t.__module__)):
        if is_defined_in_module(m.__dict__):
            return t.__module__

    module_prefix = f"{t.__module__}."
    for name, mod in sys.modules.items():
        if name.startswith(module_prefix) and is_defined_in_module(mod.__dict__):
            return name

    # it's not in the module, but keep it as a last resort, to facilitate diagnostics
    return t.__module__


RANGE_ITER_TYPE = type(iter(range(1)))

def get_type_name(obj: type, depth: int = 0) -> str:
    """Returns a type's name as a string."""

    if depth > 255:
        # We have likely fallen into an infinite recursion.
        # Fail gracefully to return "Never" while reporting the warning.
        print(f"Warning: RightTyper failed to compute the type of {obj}.")
        return "typing.Never"

    # Some builtin types are available from the "builtins" module,
    # some from the "types" module, but others still, such as
    # "list_iterator", aren't known by any particular name.
    if obj.__module__ == "builtins":
        if obj is NoneType:
            return "None"
        elif in_builtins_import(obj):
            return obj.__name__ # these are "well known", so no module name needed
        elif (name := from_types_import(obj)):
            return f"types.{name}"
        elif obj is RANGE_ITER_TYPE:
            return "typing.Iterator[int]"
        # TODO match other ABC from collections.abc based on interface
        elif issubclass(obj, abc.Iterator):
            return "typing.Iterator[typing.Any]"
        else:
            # fall back to its name, just so we can tell where it came from.
            return f"builtins.{obj.__name__}"

    # Certain dtype types' __qualname__ doesn't include a fully qualified name of their inner type
    if obj.__module__ == 'numpy' and 'dtype[' in obj.__name__ and hasattr(obj, "type"):
        t_name = obj.__qualname__.split('[')[0]
        return f"{obj.__module__}.{t_name}[{get_type_name(obj.type, depth+1)}]"

    # Disabled for now: passing types using aliases at this point can lead to
    # confusion, as there can be an alias that conflicts with a module's fully
    # qualified name.  For example, "import x.y as ast" would conflict with "ast.If"
    if False:
        # Look for a local alias for the type
        # FIXME this only checks for globally known names, and doesn't match inner classes (e.g., Foo.Bar).
        # FIXME that name may be locally bound to something different, hiding the global
        if caller_frame := find_caller_frame():
            current_namespace = caller_frame.f_globals
            if current_namespace.get(obj.__name__) is obj:
                return obj.__name__

            # Check if the type is accessible from a module in the current namespace
            for name, mod in current_namespace.items():
                if (
                    inspect.ismodule(mod)
                    and hasattr(mod, obj.__name__)
                    and getattr(mod, obj.__name__) is obj
                ):
                    return f"{name}.{obj.__name__}"

    if obj.__module__ == "__main__":    # TODO merge this into lookup_type_module
        return f"{get_main_module_fqn()}.{obj.__qualname__}"

    return f"{lookup_type_module(obj)}.{obj.__qualname__}"


def _is_instance(obj: object, types: tuple[type, ...]) -> type|None:
    """Like isinstance(), but returns the type matched."""
    for t in types:
        if isinstance(obj, t):
            return t

    return None


def get_full_type(value: Any, depth: int = 0) -> str:
    """
    get_full_type takes a value (an instance) as input and returns a string representing its type.

    If the value is a collection, it randomly selects an element (or key-value pair) and determines their types.
    If the value is a tuple, it determines the types of all elements in the tuple.

    For other types, it returns the name of the type.
    """
    if depth > 255:
        # We have likely fallen into an infinite recursion.
        # Fail gracefully to return "Never" while reporting the warning.
        print(f"Warning: RightTyper failed to compute the type of {value}.")
        return "typing.Never"

    t: type|None

    if isinstance(value, dict):
        t = type(value)
        module = "" if t.__module__ == "builtins" else f"{t.__module__}."
        if value:
            el = value.random_item() if isinstance(value, RandomDict) else sample_from_collection(value.items())
            return f"{module}{t.__qualname__}[{get_full_type(el[0], depth+1)}, {get_full_type(el[1], depth+1)}]"
        else:
            return "{module}{t.__qualname__}[typing.Never, typing.Never]"
    elif isinstance(value, (list, set)):
        t = type(value)
        module = "" if t.__module__ == "builtins" else f"{t.__module__}."
        if value:
            el = sample_from_collection(value)
            return f"{module}{t.__qualname__}[{get_full_type(el, depth+1)}]"
        else:
            return f"{module}{t.__qualname__}[typing.Never]"
    elif (t := _is_instance(value, (abc.KeysView, abc.ValuesView))):
        if value:
            el = sample_from_collection(value)
            return f"typing.{t.__qualname__}[{get_full_type(el, depth+1)}]"
        else:
            return f"typing.{t.__qualname__}[typing.Never]"
    elif isinstance(value, abc.ItemsView):
        if value:
            el = sample_from_collection(value)
            return f"typing.ItemsView[{get_full_type(el[0], depth+1)}, {get_full_type(el[1], depth+1)}]"
        else:
            return "typing.ItemsView[typing.Never, typing.Never]"
    elif isinstance(value, tuple):
        if isinstance_namedtuple(value):
            return f"{value.__class__.__module__}.{value.__class__.__qualname__}"
        else:
            if value:
                return f"tuple[{', '.join(get_full_type(elem, depth+1) for elem in value)}]"
            else:
                return "tuple"
    elif isinstance(value, (FunctionType, MethodType)):
        return type_from_annotations(value)
    elif isinstance(value, abc.Generator):
        return "typing.Generator[typing.Any, typing.Any, typing.Any]"  # FIXME needs yield / send / return types
    elif isinstance(value, abc.AsyncGenerator):
        return "typing.AsyncGenerator[typing.Any, typing.Any]"  # FIXME needs yield / send types
    elif isinstance(value, abc.Coroutine):
        return "typing.Coroutine[typing.Any, typing.Any, typing.Any]"  # FIXME needs yield / send / return types
    elif (t := type(value)).__module__ == 'numpy' and t.__qualname__ == 'ndarray':
        return f"{get_type_name(t, depth+1)}[typing.Any, {get_type_name(type(value.dtype), depth+1)}]"

    return get_type_name(type(value), depth+1)


def get_adjusted_full_type(value: Any, class_type: type|None=None) -> str:
    #print(f"{type(value)=} {class_type=}")
    if type(value) == class_type:
        return "typing.Self"

    return get_full_type(value)


def isinstance_namedtuple(obj: object) -> bool:
    return (
        isinstance(obj, tuple)
        and hasattr(obj, "_asdict")
        and hasattr(obj, "_fields")
    )


def update_argtypes(
    argtypes: list[ArgInfo],
    arg_types: dict[
        tuple[FuncInfo, ArgumentName],
        ArgumentType,
    ],
    index: tuple[FuncInfo, ArgumentName],
    arg_values: Any,
    class_type: type|None,
    arg: str,
    is_vararg: bool,
    is_kwarg: bool
) -> None:

    def add_arg_info(
        argument_name: str,
        values: Any,
        arg_type_enum: ArgumentType,
    ) -> None:
        types = TypenameSet(
            {
                TypenameFrequency(
                    Typename(get_adjusted_full_type(val, class_type)),
                    1,
                )
                for val in values
            }
        )
        argtypes.append(
            ArgInfo(
                ArgumentName(argument_name),
                types,
            )
        )
        arg_types[index] = arg_type_enum

    if is_vararg:
        add_arg_info(
            arg,
            arg_values[0],
            ArgumentType.vararg,
        )
    elif is_kwarg:
        add_arg_info(
            arg,
            arg_values[0].values(),
            ArgumentType.kwarg,
        )
    else:
        add_arg_info(
            arg,
            arg_values,
            ArgumentType.positional,
        )


def format_annotation(annotation: Any) -> str:
    """Format an annotation (type hint) as a string."""
    if isinstance(annotation, type):
        return annotation.__name__
    elif hasattr(annotation, "_name") and annotation._name is not None:
        return str(annotation._name)
    elif (
        hasattr(annotation, "__origin__") and annotation.__origin__ is not None
    ):
        origin = format_annotation(annotation.__origin__)
        args = ", ".join(
            [format_annotation(arg) for arg in annotation.__args__]
        )
        return f"{origin}[{args}]"
    else:
        return str(annotation)


def format_function_definition(
    func_name: str,
    arg_names: list[str],
    type_hints: dict[str, Any],
) -> str:
    """Format the function definition based on its name, argument names, and type hints."""
    params = []
    for arg in arg_names:
        type_hint = type_hints.get(arg)
        if type_hint:
            params.append(f"{arg}: {format_annotation(type_hint)}")
        else:
            params.append(arg)

    return_annotation = ""
    if "return" in type_hints:
        return_annotation = f" -> {format_annotation(type_hints['return'])}"

    params_str = ", ".join(params)
    function_definition = f"def {func_name}({params_str}){return_annotation}:"
    return function_definition


def get_class_source_file(cls: type) -> str:
    module_name = cls.__module__

    # Check if the class is built-in
    if module_name == "builtins":
        return ""  # Built-in classes do not have source files

    try:
        # Try to get the module from sys.modules
        module = sys.modules[module_name]
        # Try to get the __file__ attribute
        file_path = getattr(module, "__file__", None)
        if file_path:
            return str(file_path)
        # If __file__ is not available, use inspect to get the source file
        import inspect

        file_path = inspect.getfile(cls)
        return file_path
    except (KeyError, TypeError, AttributeError):
        pass

    # Derive the file path from the module name
    try:
        # Assuming the module is part of the standard package structure
        import os

        module_parts = module_name.split(".")
        file_path = os.path.join(*module_parts)
        return file_path

    except Exception:
        pass

    return ""


def _source_relative_to_pkg(file: Path) -> Path|None:
    """Returns a Python source file's path relative to its package"""
    if not file.is_absolute():
        file = file.resolve()

    parents = list(file.parents)

    for d in sys.path:
        path = Path(d)
        if not path.is_absolute():
            path = path.resolve()

        for p in parents:
            if p == path:
                return file.relative_to(p)

    return None


def source_to_module_fqn(file: Path) -> str|None:
    """Returns a source file's fully qualified package name, if possible."""
    if not (path := _source_relative_to_pkg(file)):
        return None

    path = path.parent if path.name == '__init__.py' else path.parent / path.stem
    return '.'.join(path.parts)


@cache
def get_main_module_fqn() -> str:
    main = sys.modules['__main__']
    if hasattr(main, "__file__") and main.__file__:
        if fqn := source_to_module_fqn(Path(main.__file__)):
            return fqn

    return "__main__"



import libcst as cst
import libcst.matchers as cstm
import textwrap
from righttyper.unified_transformer import UnifiedTransformer, types_in_annotation, used_names
from righttyper.righttyper_types import FuncInfo, Filename, FunctionName, Typename, ArgumentName, FuncAnnotation
import typing
import pytest
import re


def get_function(m: cst.Module, name: str) -> str|None:
    class V(cst.CSTVisitor):
        def __init__(self):
            self.found = None
            self.name_stack = []

        def visit_ClassDef(self, node: cst.ClassDef):
            self.name_stack.append(node.name.value)

        def leave_ClassDef(self, node: cst.ClassDef):
            self.name_stack.pop()

        def visit_FunctionDef(self, node: cst.FunctionDef) -> bool:
            self.name_stack.append(node.name.value)
            qual_name = ".".join(self.name_stack)
            self.name_stack.append("<locals>")
            if qual_name == name:
                self.found = node
                return False # stop here
            return True

        def leave_FunctionDef(self, node: cst.FunctionDef):
            self.name_stack.pop()
            self.name_stack.pop()

    v = V()
    m.visit(v)
    return cst.Module([v.found]).code.lstrip('\n') if v.found else None


def assert_regex(pattern: str, text: str|None) -> re.Match:
    m = re.search(pattern, text if text else "")
    assert m, f"'{text}' doesn't match '{pattern}'"
    return m


def get_if_type_checking(m: cst.Module) -> str|None:
    stmts = typing.cast(list[cst.If], list(cstm.findall(m, cstm.If(test=cstm.Name("TYPE_CHECKING")))))
    return cst.Module(stmts).code.lstrip('\n') if stmts else None


def _split(s: str) -> tuple[str, str]:
    parts = s.split('.')
    return parts[0], '.'.join(parts[1:])


def test_transform_function():
    code = cst.parse_module(textwrap.dedent("""\
        def foo(x, y):
            return (x+y)/2

        def bar(x):
            return x/2

        def baz(z):
            return z/2
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    baz = FuncInfo(Filename('foo.py'), FunctionName('baz'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('int'))
                    ],
                    Typename('float')
                ),
                baz: FuncAnnotation(
                    [
                        (ArgumentName('z'), Typename('int'))
                    ],
                    None
                )
            },
            override_annotations=False,
            module_name='foo',
            module_names=['foo']
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: int, y) -> float:
            return (x+y)/2
    """)

    assert get_function(code, 'bar') == textwrap.dedent("""\
        def bar(x):
            return x/2
    """)

    assert get_function(code, 'baz') == textwrap.dedent("""\
        def baz(z: int):
            return z/2
    """)

    assert get_if_type_checking(code) == None


def test_transform_method():
    code = cst.parse_module(textwrap.dedent("""\
        class C:
            def foo(self, x, y):
                return (x+y)/2

            @staticmethod
            def bar(x):
                return x/2

            @classmethod
            def baz(cls, z):
                return z/2
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('C.foo'))
    bar = FuncInfo(Filename('foo.py'), FunctionName('C.bar'))
    baz = FuncInfo(Filename('foo.py'), FunctionName('C.baz'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('int'))
                    ],
                    Typename('float')
                ),
                bar: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('int'))
                    ],
                    Typename('float')
                ),
                baz: FuncAnnotation(
                    [
                        (ArgumentName('z'), Typename('int'))
                    ],
                    Typename('float')
                )
            },
            override_annotations=False,
            module_name='foo',
            module_names=['foo']
        )

    code = code.visit(t)
    assert get_function(code, 'C.foo') == textwrap.dedent("""\
        def foo(self, x: int, y) -> float:
            return (x+y)/2
    """)

    assert get_function(code, 'C.bar') == textwrap.dedent("""\
        @staticmethod
        def bar(x: int) -> float:
            return x/2
    """)

    assert get_function(code, 'C.baz') == textwrap.dedent("""\
        @classmethod
        def baz(cls, z: int) -> float:
            return z/2
    """)

    assert get_if_type_checking(code) == None


def test_transform_local_function():
    code = cst.parse_module(textwrap.dedent("""\
        def foo(x, y):
            def bar(z):
                return z/2
            return bar(x+y)
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    bar = FuncInfo(Filename('foo.py'), FunctionName('foo.<locals>.bar'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('int')),
                        (ArgumentName('y'), Typename('float'))
                    ],
                    Typename('float')
                ),
                bar: FuncAnnotation(
                    [
                        (ArgumentName('z'), Typename('int'))
                    ],
                    Typename('float')
                ),
            },
            override_annotations=False,
            module_name='foo',
            module_names=['foo']
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: int, y: float) -> float:
            def bar(z: int) -> float:
                return z/2
            return bar(x+y)
    """)

    assert get_if_type_checking(code) == None


def test_override_annotations():
    code = cst.parse_module(textwrap.dedent("""\
        def foo(x: int) -> int:
            return x/2

        class C:
            def bar(self, x: float) -> int:
                return x/2
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    bar = FuncInfo(Filename('foo.py'), FunctionName('C.bar'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('float'))
                    ],
                    Typename('float')
                ),
                bar: FuncAnnotation(
                    [
                        (ArgumentName('self'), Typename('typing.Self')),
                        (ArgumentName('x'), Typename('int'))
                    ],
                    Typename('float')
                ),
            },
            override_annotations=True,
            module_name='foo',
            module_names=['foo']
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: float) -> float:
            return x/2
    """)

    assert get_function(code, 'C.bar') == textwrap.dedent("""\
        def bar(self: Self, x: int) -> float:
            return x/2
    """)


def test_transform_adds_typing_import_for_typing_names():
    code = cst.parse_module(textwrap.dedent("""\
        def foo(x): ...
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('typing.Optional[int]'))
                    ],
                    Typename('list[typing.Never]')
                )
            },
            override_annotations=False,
            module_name='foo',
            module_names=['foo']
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: Optional[int]) -> list[Never]: ...
    """)

    code_str = str(code.code)
    assert re.search(r"^ *from typing import .*\bOptional\b", code_str)
    assert re.search(r"^ *from typing import .*\bNever\b", code_str)
    assert get_if_type_checking(code) == None


def test_transform_unknown_type_as_string():
    code = cst.parse_module(textwrap.dedent("""\
        def foo(x, y):
            return x/2
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('int')),
                        (ArgumentName('y'), Typename('x.y.WholeNumber|None'))
                    ],
                    Typename('x.z.FloatingPointNumber')
                )
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                "foo",
                "x.y",
                "x"
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: int, y: "x.y.WholeNumber|None") -> "x.z.FloatingPointNumber":
            return x/2
    """)

    assert get_if_type_checking(code) == textwrap.dedent("""\
        if TYPE_CHECKING:
            import x
            import x.y
    """)


def test_transform_unknown_type_with_import_annotations():
    code = cst.parse_module(textwrap.dedent("""\
        from __future__ import annotations

        def foo(x, y):
            return x/2
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('int')),
                        (ArgumentName('y'), Typename('x.y.WholeNumber|None'))
                    ],
                    Typename('x.z.FloatingPointNumber')
                )
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                "foo",
                "x.y",
                "x"
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: int, y: x.y.WholeNumber|None) -> x.z.FloatingPointNumber:
            return x/2
    """)

    assert get_if_type_checking(code) == textwrap.dedent("""\
        if TYPE_CHECKING:
            import x
            import x.y
    """)

    code_str = str(code.code).strip()
    assert code_str.startswith("from __future__ import annotations")


def test_transform_deletes_type_hint_comments_in_header():
    code = cst.parse_module(textwrap.dedent("""\
        def foo(x, y): # type: (int, int) -> Any
            return (x+y)/2

        def bar(x):   # type: (Any) -> None
            pass
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('int')),
                        (ArgumentName('y'), Typename('int'))
                    ],
                    None
                )
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo'
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: int, y: int):
            return (x+y)/2
    """)

    assert get_function(code, 'bar') == textwrap.dedent("""\
        def bar(x):   # type: (Any) -> None
            pass
    """)

    assert get_if_type_checking(code) == None


def test_transform_deletes_type_hint_comments_in_parameters():
    code = cst.parse_module(textwrap.dedent("""\
        def foo(
            x,  # type: int
            y   # type: float
        ):
            # type: (...) -> Any
            return (x+y)/2

        def bar(
            x   # type: Any
        ):
            # type: (...) -> None
            pass
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('int')),
                        (ArgumentName('y'), Typename('int'))
                    ],
                    None
                )
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo'
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(
            x: int,
            y: int
        ):
            # type: (...) -> Any
            return (x+y)/2
    """)

    assert get_function(code, 'bar') == textwrap.dedent("""\
        def bar(
            x   # type: Any
        ):
            # type: (...) -> None
            pass
    """)

    assert get_if_type_checking(code) == None


def test_transform_deletes_type_hint_comments_for_retval():
    code = cst.parse_module(textwrap.dedent("""\
        def foo(
            x,  # type: int
            y   # type: float
        ):
            # type: (...) -> Any
            return (x+y)/2

        def bar(
            x   # type: Any
        ):
            # type: (...) -> None
            pass
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                    ],
                    Typename('float')
                )
            },
            override_annotations=False,
            module_name = 'foo',
            module_names=[
                'foo'
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(
            x,  # type: int
            y   # type: float
        ) -> float:
            return (x+y)/2
    """)

    assert get_function(code, 'bar') == textwrap.dedent("""\
        def bar(
            x   # type: Any
        ):
            # type: (...) -> None
            pass
    """)

    assert get_if_type_checking(code) == None


def test_transform_locally_defined_types():
    code = cst.parse_module(textwrap.dedent("""\
        def foo(x, y):
            return F((x+y)/2)

        class F:
            def __init__(self, v):
                self.v = v

            def foo(self, v):
                return F(v)

        def bar(x, y):
            return F((x+y)/2)
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    f_foo = FuncInfo(Filename('foo.py'), FunctionName('F.foo'))
    bar = FuncInfo(Filename('foo.py'), FunctionName('bar'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('int')),
                        (ArgumentName('y'), Typename('int'))
                    ],
                    Typename('foo.F')
                ),
                f_foo: FuncAnnotation(
                    [
                        (ArgumentName('v'), Typename('float')),
                    ],
                    Typename('foo.F')
                ),
                bar: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('int')),
                        (ArgumentName('y'), Typename('int'))
                    ],
                    Typename('foo.F')
                )
            },
            override_annotations=False,
            module_name = 'foo',
            module_names=[
                'foo'
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: int, y: int) -> "F":
            return F((x+y)/2)
    """)

    assert get_function(code, 'F.foo') == textwrap.dedent("""\
        def foo(self, v: float) -> "F":
            return F(v)
    """)

    # F is now known, so no quotes are needed
    assert get_function(code, 'bar') == textwrap.dedent("""\
        def bar(x: int, y: int) -> F:
            return F((x+y)/2)
    """)

    assert get_if_type_checking(code) == None


def test_uses_imported_aliases():
    code = cst.parse_module(textwrap.dedent("""\
        from x.y import z as zed
        from y import T
        if True:
            import a.b as A

        def foo(x, y, z): ...

        import r    # imported after 'def foo', so can't be used in annotation
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('x.y.z')),
                        (ArgumentName('y'), Typename('y.T')),
                        (ArgumentName('z'), Typename('a.b.c.T')),
                    ],
                    Typename('r.t.T')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'x',
                'x.y',
                'y',
                'a',
                'a.b',
                'r'
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: zed, y: T, z: A.c.T) -> "r.t.T": ...
    """)


    assert get_if_type_checking(code) == textwrap.dedent("""\
        if TYPE_CHECKING:
            import r
    """)


def test_uses_imported_domains():
    code = cst.parse_module(textwrap.dedent("""\
        if True:
            import x.y

        with something():
            import a

        def foo(x, y): ...

        import r    # imported after 'def foo', so can't be used in annotation
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('x.y.z')),
                        (ArgumentName('y'), Typename('a.T')),
                    ],
                    Typename('r.t.T')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'a',
                'x',
                'x.y',
                'r'
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: x.y.z, y: a.T) -> "r.t.T": ...
    """)

    assert get_if_type_checking(code) == textwrap.dedent("""\
        if TYPE_CHECKING:
            import r
    """)


def test_imports_subdomain_if_needed():
    code = cst.parse_module(textwrap.dedent("""\
        import x
        import a.b.c

        def foo(x): ...
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('x.y.z')),
                    ],
                    Typename('a.b')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'x',
                'x.y',
                'a',
                'a.b',
                'a.b.c',
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: "x.y.z") -> a.b: ...
    """)

    assert get_if_type_checking(code) == textwrap.dedent("""\
        if TYPE_CHECKING:
            import x.y
    """)


def test_existing_typing_imports():
    code = cst.parse_module(textwrap.dedent("""\
        from typing import TYPE_CHECKING
        if TYPE_CHECKING:
            import m
            import ast

        def foo(x): ...

        def bar(x: "ast.For") -> "m.T": ...
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('ast.If')),
                    ],
                    Typename('typing.Any')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'ast',
                'm',
                'typing',
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: "ast.If") -> Any: ...
    """)

    assert get_if_type_checking(code) == textwrap.dedent("""\
        if TYPE_CHECKING:
            import m
            import ast
            import ast
    """)

    code_str = str(code.code)
    assert code_str.startswith(textwrap.dedent("""\
        from typing import TYPE_CHECKING, Any
        if TYPE_CHECKING:
            import m
    """))


def test_inserts_imports_after_docstring_and_space():
    code = cst.parse_module(textwrap.dedent("""\
        '''blah blah blah
        blah
        '''


        from __future__ import annotations

        def foo(x): ...
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('ast.If')),
                    ],
                    Typename('typing.Any')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'ast',
                'typing',
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: ast.If) -> Any: ...
    """)

    print(code.code)

    assert get_if_type_checking(code) == textwrap.dedent("""\
        if TYPE_CHECKING:
            import ast
    """)

    code_str = str(code.code)
    assert code_str.startswith(textwrap.dedent("""\
        '''blah blah blah
        blah
        '''


        from __future__ import annotations
        from typing import TYPE_CHECKING, Any
        if TYPE_CHECKING:
            import ast
    """))


def test_relative_import():
    code = cst.parse_module(textwrap.dedent("""\
        from .. import b
        from . import c
        from .c import X

        def foo(x, y, z): ...
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('pkg.b.T')),
                        (ArgumentName('y'), Typename('pkg.a.c.T')),
                        (ArgumentName('z'), Typename('pkg.a.c.X')),
                    ],
                    None
                ),
            },
            override_annotations=False,
            module_name = 'pkg.a.a',
            module_names = [
                'pkg.a',
                'pkg.a.a',
                'pkg.b',
                'pkg.a.c',
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: b.T, y: c.T, z: X): ...
    """)

    assert get_if_type_checking(code) == None


@pytest.mark.skip(reason="Not yet supported")
def test_uses_local_imports():
    code = cst.parse_module(textwrap.dedent("""\
        def foo():
            import m.n
            def bar(x): ...

        class C:
            import n
            def foo(self, x): ...

            class D:
                def foo(self, x): ...

        def f(a, b): ...
    """))

    foobar = FuncInfo(Filename('foo.py'), FunctionName('foo.<locals>.bar'))
    Cfoo = FuncInfo(Filename('foo.py'), FunctionName('C.foo'))
    Dfoo = FuncInfo(Filename('foo.py'), FunctionName('C.D.foo'))
    f = FuncInfo(Filename('foo.py'), FunctionName('f'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foobar: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('m.n.T')),
                    ],
                    Typename('None')
                ),
                Cfoo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('n.o.T')),
                    ],
                    Typename('None')
                ),
                Dfoo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('n.o.T')),
                    ],
                    Typename('None')
                ),
                f : FuncAnnotation(
                    [
                        (ArgumentName('a'), Typename('m.n.T')),
                        (ArgumentName('b'), Typename('n.o.T')),
                    ],
                    Typename('None')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'm.n',
                'n',
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo.<locals>.bar') == textwrap.dedent("""\
        def bar(x: m.n.T): ...
    """)

    assert get_function(code, 'C.foo') == textwrap.dedent("""\
        def foo(self, x: n.o.T): ...
    """)
    assert get_function(code, 'C.D.foo') == textwrap.dedent("""\
        def foo(self, x: "n.o.T"): ...
    """)
    assert get_function(code, 'f') == textwrap.dedent("""\
        def f(a: "m.n.T", b: "n.o.T"): ...
    """)

    assert get_if_type_checking(code) == textwrap.dedent("""\
        if TYPE_CHECKING:
            import m.n
            import n
    """)


def test_nonglobal_imported_modules_are_ignored():
    code = cst.parse_module(textwrap.dedent("""\
        def foo(x, y):
            import a.b

        class C:
            import m as m2

        def bar(x):
            pass
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    bar = FuncInfo(Filename('foo.py'), FunctionName('bar'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('a.T')),
                        (ArgumentName('y'), Typename('a.b.T')),
                    ],
                    Typename('a.c.T')
                ),
                bar: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('m.T')),
                    ],
                    Typename('m.T')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'a',
                'a.b',
                'a.c',
                'm',
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: "a.T", y: "a.b.T") -> "a.c.T":
            import a.b
    """)

    assert get_function(code, 'bar') == textwrap.dedent("""\
        def bar(x: "m.T") -> "m.T":
            pass
    """)

    assert get_if_type_checking(code) == textwrap.dedent("""\
        if TYPE_CHECKING:
            import a
            import a.b
            import a.c
            import m
    """)


def test_nonglobal_assignments_are_ignored():
    code = cst.parse_module(textwrap.dedent("""\
        from typing import Any

        def foo(x):
            a = Any

        class C:
            m = Any

        def bar(x):
            pass
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    bar = FuncInfo(Filename('foo.py'), FunctionName('bar'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('a.T')),
                    ],
                    Typename('')
                ),
                bar: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('m.T')),
                    ],
                    Typename('')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'a',
                'm'
            ]
        )

    code = code.visit(t)
    assert get_function(code, 'foo') == textwrap.dedent("""\
        def foo(x: "a.T"):
            a = Any
    """)

    assert get_function(code, 'bar') == textwrap.dedent("""\
        def bar(x: "m.T"):
            pass
    """)

    assert get_if_type_checking(code) == textwrap.dedent("""\
        if TYPE_CHECKING:
            import a
            import m
    """)


def test_if_type_checking_insertion():
    code = cst.parse_module(textwrap.dedent("""\
        from a import b
        from typing import Any

        def foo(x): ...
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('c.T')),
                    ],
                    Typename('None')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'a',
                'c',
                'typing'
            ]
        )

    code = code.visit(t)

    assert str(code.code).startswith(textwrap.dedent("""\
        from typing import TYPE_CHECKING, Any
        if TYPE_CHECKING:
            import c
    """))


def test_import_conflicts_with_import():
    code = cst.parse_module(textwrap.dedent("""\
        from a import b as a
        import b as c

        def foo(x, y): ...
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('a.T')),
                        (ArgumentName('y'), Typename('c.d.e.T')),
                    ],
                    None
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'a',
                'b',
                'c',
                'c.d'
            ]
        )

    code = code.visit(t)

    m = assert_regex(r'def foo\(x: "(.*?)", y: "(.*?)"\): ...', get_function(code, 'foo'))
    m1, t1 = _split(m.group(1))
    m2, t2 = _split(m.group(2))

    assert m1 != 'a'
    assert t1 == 'T'

    assert m2 != 'b'
    assert t2 == 'e.T'

    m1, m2 = sorted((m1, m2))

    print(code.code)

    assert get_if_type_checking(code) == textwrap.dedent(f"""\
        if TYPE_CHECKING:
            import a as {m1}
            import c.d as {m2}
    """)


def test_import_conflicts_with_definitions():
    code = cst.parse_module(textwrap.dedent("""\
        class a: ...
        def c(): pass

        def foo(x, y): ...
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('a.T')),
                        (ArgumentName('y'), Typename('c.d.e.T')),
                    ],
                    None
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'a',
                'b',
                'c',
                'c.d'
            ]
        )

    code = code.visit(t)

    m = assert_regex(r'def foo\(x: "(.*?)", y: "(.*?)"\): ...', get_function(code, 'foo'))
    m1, t1 = _split(m.group(1))
    m2, t2 = _split(m.group(2))

    assert m1 != 'a'
    assert t1 == 'T'

    assert m2 != 'b'
    assert t2 == 'e.T'

    m1, m2 = sorted((m1, m2))

    print(code.code)

    assert get_if_type_checking(code) == textwrap.dedent(f"""\
        if TYPE_CHECKING:
            import a as {m1}
            import c.d as {m2}
    """)


def test_import_conflicts_with_assignments():
    code = cst.parse_module(textwrap.dedent("""\
        a, b = (10, 20)

        def foo(x, y): ...

        c: int = 10
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('a.T')),
                        (ArgumentName('y'), Typename('c.d.e.T')),
                    ],
                    None
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'a',
                'b',
                'c',
                'c.d'
            ]
        )

    code = code.visit(t)

    m = assert_regex(r'def foo\(x: "(.*?)", y: "(.*?)"\): ...', get_function(code, 'foo'))
    m1, t1 = _split(m.group(1))
    m2, t2 = _split(m.group(2))

    assert m1 != 'a'
    assert t1 == 'T'

    assert m2 != 'b'
    assert t2 == 'e.T'

    m1, m2 = sorted((m1, m2))

    print(code.code)

    assert get_if_type_checking(code) == textwrap.dedent(f"""\
        if TYPE_CHECKING:
            import a as {m1}
            import c.d as {m2}
    """)


def test_import_conflicts_with_with():
    code = cst.parse_module(textwrap.dedent("""\
        with my_handler() as a:
            pass

        def foo(x): ...
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('a.T')),
                    ],
                    None
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'a',
            ]
        )

    code = code.visit(t)

    m = assert_regex(r'def foo\(x: "(.*?)"\): ...', get_function(code, 'foo'))
    m1, t1 = _split(m.group(1))

    assert m1 != 'a'
    assert t1 == 'T'

    print(code.code)

    assert get_if_type_checking(code) == textwrap.dedent(f"""\
        if TYPE_CHECKING:
            import a as {m1}
    """)


def test_import_conflicts_alias_for_module():
    code = cst.parse_module(textwrap.dedent("""\
        a, b = (10, 20)

        def foo(x): ...
    """))

    foo = FuncInfo(Filename('foo.py'), FunctionName('foo'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                foo: FuncAnnotation(
                    [
                        (ArgumentName('x'), Typename('a')), # module "a" meant here, not something in it
                    ],
                    None
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'a',
            ]
        )

    code = code.visit(t)

    m = assert_regex(r'def foo\(x: "(.*?)"\): ...', get_function(code, 'foo'))
    m1, t1 = _split(m.group(1))

    assert m1 != 'a'
    assert t1 == ''

    print(code.code)

    assert get_if_type_checking(code) == textwrap.dedent(f"""\
        if TYPE_CHECKING:
            import a as {m1}
    """)


def test_builtin_name_conflicts():
    code = cst.parse_module(textwrap.dedent("""\
    def int():
        pass

    class C:
        @property
        def tuple(self):
            pass

        def f(self):
            pass
    """))

    f = FuncInfo(Filename('foo.py'), FunctionName('C.f'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                f: FuncAnnotation(
                    [
                    ],
                    Typename('tuple[int, float]')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'builtins',
            ]
        )

    code = code.visit(t)

    assert get_function(code, 'C.f') == textwrap.dedent("""\
        def f(self) -> "builtins.tuple[builtins.int, float]":
            pass
    """)

    assert get_if_type_checking(code) == textwrap.dedent(f"""\
        if TYPE_CHECKING:
            import builtins
    """)


def test_class_names_dont_affect_body_of_methods():
    code = cst.parse_module(textwrap.dedent("""\
    tuple = 0

    class C:
        int = 0 # doesn't affect 'g'

        def f(self):
            def g(x):
                pass
    """))

    g = FuncInfo(Filename('foo.py'), FunctionName('C.f.<locals>.g'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                g: FuncAnnotation(
                    [
                    ],
                    Typename('tuple[int]')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'builtins',
            ]
        )

    code = code.visit(t)

    assert get_function(code, 'C.f.<locals>.g') == textwrap.dedent("""\
        def g(x) -> "builtins.tuple[int]":
            pass
    """)

    assert get_if_type_checking(code) == textwrap.dedent(f"""\
        if TYPE_CHECKING:
            import builtins
    """)


def test_inner_function():
    code = cst.parse_module(textwrap.dedent("""\
    tuple = 0

    class C:
        def f(self):
            int = 0

            def g(x):
                def h(x):
                    pass

            class D:
                float = 0

                def i(self):
                    def j(x):
                        pass
    """))

    g = FuncInfo(Filename('foo.py'), FunctionName('C.f.<locals>.g'))
    h = FuncInfo(Filename('foo.py'), FunctionName('C.f.<locals>.g.<locals>.h'))
    i = FuncInfo(Filename('foo.py'), FunctionName('C.f.<locals>.D.i'))
    j = FuncInfo(Filename('foo.py'), FunctionName('C.f.<locals>.D.i.<locals>.j'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                g: FuncAnnotation([], Typename('tuple[int, float]')),
                h: FuncAnnotation([], Typename('tuple[int, float]')),
                i: FuncAnnotation([], Typename('tuple[int, float]')),
                j: FuncAnnotation([], Typename('tuple[int, float]')),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'builtins',
            ]
        )

    code = code.visit(t)

    assert get_function(code, 'C.f.<locals>.g') == textwrap.dedent("""\
        def g(x) -> "builtins.tuple[builtins.int, float]":
            def h(x) -> "builtins.tuple[builtins.int, float]":
                pass
    """)

    assert get_function(code, 'C.f.<locals>.D.i') == textwrap.dedent("""\
        def i(self) -> "builtins.tuple[builtins.int, builtins.float]":
            def j(x) -> "builtins.tuple[builtins.int, float]":
                pass
    """)

    assert get_if_type_checking(code) == textwrap.dedent(f"""\
        if TYPE_CHECKING:
            import builtins
    """)


def test_builtin_name_conflicts_even_module_name():
    code = cst.parse_module(textwrap.dedent("""\
    builtins = 'even this!'

    def int():
        pass

    class C:
        @property
        def tuple(self):
            pass

        def f(self):
            pass
    """))

    f = FuncInfo(Filename('foo.py'), FunctionName('C.f'))
    t = UnifiedTransformer(
            filename='foo.py',
            type_annotations = {
                f: FuncAnnotation(
                    [
                    ],
                    Typename('tuple[int, float]')
                ),
            },
            override_annotations=False,
            module_name = 'foo',
            module_names = [
                'foo',
                'builtins',
            ]
        )

    code = code.visit(t)

    m = assert_regex(r'def f\(self\) -> \"(.*?)\":', get_function(code, 'C.f'))
    types = types_in_annotation(cst.parse_expression(m.group(1)))

    assert 'float' in types # no conflict

    for typ in types - {'float'}:
        assert typ not in ('int', 'tuple'), f"{typ=}"

        mod, rest = _split(typ)
        assert mod not in ('', 'builtins'), f"{typ=}"
        assert rest in ('int', 'tuple'), f"{typ=}"

    assert get_if_type_checking(code) == textwrap.dedent(f"""\
        if TYPE_CHECKING:
            import builtins as {mod}
    """)


def test_used_names():
    code = cst.parse_module(textwrap.dedent("""\
    a, b = 0, 0
    c: int = (d := 0)

    class C:
        e = 0

        class D:
            def f(self):
                pass

        def g(self):
            h = 0
            def i(self): pass

    def j():
        pass

    with handler as (k,):
        pass

    with handler as [l]:
        pass
    """))

    assert {'a', 'b', 'c', 'd', 'C', 'j', 'k', 'l'} == used_names(code)

    C = typing.cast(cst.ClassDef, cstm.findall(code, cstm.ClassDef(name=cstm.Name('C')))[0])
    assert {'D', 'e', 'g'} == used_names(C)

    D = typing.cast(cst.ClassDef, cstm.findall(code, cstm.ClassDef(name=cstm.Name('D')))[0])
    assert {'f'} == used_names(D)

    f = typing.cast(cst.FunctionDef, cstm.findall(code, cstm.FunctionDef(name=cstm.Name('f')))[0])
    assert set() == used_names(f)

    g = typing.cast(cst.FunctionDef, cstm.findall(code, cstm.FunctionDef(name=cstm.Name('g')))[0])
    assert {'h', 'i'} == used_names(g)

    i = typing.cast(cst.FunctionDef, cstm.findall(code, cstm.FunctionDef(name=cstm.Name('i')))[0])
    assert set() == used_names(i)


def test_types_in_annotation():
    def get_types(s):
        return types_in_annotation(cst.parse_expression(s))

    assert {'int'} == get_types('int')
    assert {'Tuple', 'int', 'float'} == get_types('Tuple[int, float]')
    assert {'Dict', 'foo.bar', 'bar.baz'} == get_types('Dict[foo.bar, bar.baz]')
    assert {'Union', 'int', 'float', 'Tuple', 'a.b.c'} == get_types('Union[int, float, Tuple[int, a.b.c]]')



import textwrap
import subprocess
import sys
from pathlib import Path


def test_issue_22(tmp_path, monkeypatch):
    t = textwrap.dedent("""\
        def extracted_function(A):
            return all(A[i] <= A[i + 1] for i in range(len(A) - 1)) or all(A[i] >=
                A[i + 1] for i in range(len(A) - 1))

        def optimized(A):
            return all(A[i] <= A[i + 1] for i in range(len(A) - 1)) and all(A[i] >=
                A[i + 1] for i in range(len(A) - 1))

        def main():
            assert extracted_function([6, 5, 4, 4]) == True
            assert extracted_function([1, 2, 2, 3]) == True
            assert extracted_function([1, 3, 2]) == False

        if __name__ == '__main__':
            main()
        """)

    monkeypatch.chdir(tmp_path)
    Path("t.py").write_text(t)

    subprocess.run([sys.executable, '-m', 'righttyper', '--overwrite', '--output-files', 't.py'])
    
    assert "def extracted_function(A: list[int]) -> bool" in Path("t.py").read_text()



import json
from collections import defaultdict

import click
from rich.console import Console
from rich.table import Table

from typing import TypedDict


class Error(TypedDict):
    severity: str
    code: str


# Define the error categories
CATEGORIES: dict[str, str] = {
    "attr-defined": "Attribute Access: Attempting to access an attribute that doesn't exist",
    "arg-type": "Argument Type: Function argument has an incompatible type",
    "call-arg": "Call Argument: Invalid arguments passed to a function or method",
    "call-overload": "Call Overload: No overload matches the arguments passed",
    "assignment": "Assignment: Incompatible types in assignment",
    "index": "Indexing: Indexing operation on an invalid type",
    "name-defined": "Name Error: Name is not defined or used before assignment",
    "method-assign": "Method Assignment: Cannot assign to a method",
    "overload-cannot-match": "Overload Cannot Match: function signature will never be matched",
    "operator": "Operator: Unsupported operation for the types involved",
    "return-value": "Return Value: Return type is incompatible with the declared type",
    "union-attr": "Union Attribute: Attribute access on a union type not available in all members",
    "unreachable": "Unreachable Code: Code detected that is never executed",
    "misc": "Miscellaneous: General errors that don't fall into other categories",
    "attr-assign": "Attribute Assignment: Assignment of an attribute with an incompatible type",
    "assignment-none": "Assignment None: Assigning None to a non-optional type",
    "attr-assignment": "Attribute Assignment: Invalid assignment to an attribute",
    "redundant-cast": "Redundant Cast: Casting to the same type or redundant type casting",
    "redundant-expr": "Redundant Expression: Expression doesn't affect the outcome",
    "redundant-isinstance": "Redundant isinstance: isinstance check that is always true or false",
    "redundant-typevar": "Redundant TypeVar: Type variable usage is redundant",
    "import-untyped": "Import Untyped: Imported module is not typed",
    "callable-arg": "Callable Argument: Incorrect number or type of arguments passed to a callable",
    "var-annotated": "Variable Annotations: Variable has a wrong or missing type annotation",
    "var-missing": "Variable Missing: A variable is missing or undefined",
    "truthy-bool": "Boolean Context: Object in a boolean context without a clear truth value",
    "type-arg": "Type Argument: Invalid type argument provided",
    "type-var": "Type Variable: Misuse of a type variable",
    "unsubscriptable": "Unsubscriptable: Attempt to subscript a non-subscriptable type",
    "check-untyped-def": "Untyped Function: Function has no type annotations",
    "valid-type": "Type Validity: Invalid type annotation or comment",
    "var-no-assigned": "Variable Not Assigned: Variable declared but not assigned a value",
    "var-incompatible": "Variable Incompatible: Variable assigned with an incompatible type",
    "var-no-annotated": "Variable No Annotated: Missing type annotation where required",
    "var-implicit-none": "Implicit None: Variable is implicitly None potentially leading to errors",
    "redundant-operation": "Redundant Operation: Unnecessary operations like adding zero",
    "return-incompatible": "Return Incompatible: Return value doesn't match the function's return type",
    "return-not-required": "Return Not Required: Return statement in a function that doesn't need one",
    "shadowed": "Shadowed Name: A name is shadowed by another name in the same scope",
    "star-import": "Star Import: Star imports (e.g. `from module import *`) are discouraged",
    "str-bytes-concat": "String Bytes Concatenation: Concatenating a string with bytes is not allowed",
    "syntax": "Syntax: General syntax errors detected",
    "too-many-locals": "Too Many Locals: Function has too many local variables",
    "too-many-returns": "Too Many Returns: Function has too many return statements",
    "too-many-statements": "Too Many Statements: Function has too many statements; consider refactoring",
    "tuple-index": "Tuple Index: Invalid index operation on a tuple",
    "tuple-item": "Tuple Item: Invalid tuple item access",
    "used-before-definition": "Used Before Definition: Variable used before it was defined",
    "unused-ignore": "Unused Ignore: `# type: ignore` comment is unnecessary",
    "untyped-attr": "Untyped Attribute: Attribute access on an untyped object",
    "untyped-def": "Untyped Function Definition: Function or method lacks type annotations",
    "untyped-exception": "Untyped Exception: Raising or handling an exception without typing",
    "untyped-call": "Untyped Call: Calling a function or method without type annotations",
    "var-shadowed": "Variable Shadowed: Variable is shadowed by another variable in the same scope",
    "yield-from": "Yield From: Incorrect use of the `yield from` statement",
    "yield-expected": "Yield Expected: Missing yield statement in a generator function",
    "literal-required": "Literal Required: Non-literal type provided where a literal is required",
    "literal-comparison": "Literal Comparison: Unsafe or incompatible literal comparison",
    "super-init": "Super Init: Issues with calling the superclass `__init__` method",
    "redundant-else": "Redundant Else: Else clause is unnecessary because conditions are exhaustive",
    "redundant-typevar-bound": "Redundant TypeVar Bound: TypeVar bounds are redundant or unnecessary",
    "redundant-optional": "Redundant Optional: Optional type is redundant",
    "redundant-f-string": "Redundant f-string: f-string usage where a regular string would suffice",
    "simplifiable-if": "Simplifiable If: If statement can be simplified",
    "simplifiable-union": "Simplifiable Union: Union type can be simplified",
    "missing-type-var": "Missing TypeVar: TypeVar missing in function or class",
    "overload-impl": "Overload Implementation: Implementation doesn't match overload signatures",
    "var-hint-comment": "Variable Hint Comment: Issues with type hints provided as comments",
    "no-redef": "Redefinition: A name is redefined in an incompatible way",
    "str-format": "String Format: Issues with string formatting or interpolation",
    "yield-value": "Yield Value: Issues with values yielded by a generator function",
    "literal-bounds": "Literal Bounds: Literal value out of bounds",
    "dict-arg": "Dictionary Argument: Invalid arguments passed to a dictionary operation",
    "dict-item": "Dictionary Item: Access or assignment of a dictionary item with an incompatible type",
    "list-item": "List Item: Access or assignment of a list item with an incompatible type",
    "set-item": "Set Item: Access or assignment of a set item with an incompatible type",
    "none-return": "None Return: Returning `None` where a non-None type is expected",
    "none-arg": "None Argument: Passing `None` where a non-None type is expected",
    "import-not-found": "Import Not Found: Missing or uninstalled import",
    "override": "Override: Subclass method violates Liskov substitutability principle",
    "func-returns-value": "Function Returns Value: Function returns `None` but a value is expected",
    "has-type": "Has Type: Cannot determine type of a variable",
    "return": "Return: Missing return statement",
    "name-match": "Name Match: Argument type mismatch in a namedtuple",
    "truthy-function": "Truthy Function: Function is always true in a boolean context (conditional)",
}


def classify_errors(errors: list[Error]) -> dict[str, list[Error]]:
    classified_errors: dict[str, list[Error]] = defaultdict(list)
    for error in errors:
        if error["severity"] != "error":
            continue
        category = CATEGORIES.get(error["code"], "Unknown")
        if category == "Unknown":
            print(f"Unknown: {error}")
        classified_errors[category].append(error)
    return classified_errors


def summarize_errors(
    classified_errors: dict[str, list[Error]]
) -> tuple[list[tuple[str, str, int, str]], int]:
    total_errors = sum(len(errors) for errors in classified_errors.values())
    summary: list[tuple[str, str, int, str]] = []

    for category, errors in classified_errors.items():
        count = len(errors)
        percentage = (count / total_errors) * 100
        summary.append(
            (category, f"{errors[0]['code']}", count, f"{percentage:.2f}%")
        )

    # Sort by percentage in descending order
    summary.sort(key=lambda x: float(x[3][:-1]), reverse=True)

    return summary, total_errors


def display_summary(
    summary: list[tuple[str, str, int, str]], total_errors: int
) -> None:
    console = Console()
    table = Table(title="Mypy Error Summary")
    table.add_column("Category", justify="left", style="cyan", no_wrap=True)
    table.add_column("Error Code", justify="left", style="cyan")
    table.add_column("Count", justify="right", style="magenta")
    table.add_column("Percentage", justify="right", style="green")

    for category, error_code, count, percentage in summary:
        table.add_row(
            category, f"[bold]{error_code}[/bold]", str(count), percentage
        )

    console.print(table)
    console.print(f"\nTotal Errors: {total_errors}\n")


def display_summary_markdown(
    summary: list[tuple[str, str, int, str]], total_errors: int
) -> None:
    # Generate markdown table
    markdown_table = "# Mypy Error Summary\n\n"
    markdown_table += "| Category | Error Code | Count | Percentage |\n"
    markdown_table += "|---|---|---|---|\n"

    for category, error_code, count, percentage in summary:
        markdown_table += (
            f"| {category} | {error_code} | {count} | {percentage} |\n"
        )

    markdown_table += f"\n**Total Errors**: {total_errors}\n"

    print(markdown_table)


def display_summary_csv(
    summary: list[tuple[str, str, int, str]], total_errors: int
) -> None:
    # Generate CSV-style output
    csv_table = "Category,Error Code,Count,Percentage\n"

    for category, error_code, count, percentage in summary:
        csv_table += f"{category},{error_code},{count},{percentage}\n"

    csv_table += f"\nTotal Errors,{total_errors},\n"
    print(csv_table)


@click.command()
@click.option("--markdown", is_flag=True, default=False)
@click.option("--csv", is_flag=True, default=False)
@click.argument(
    "file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    required=True,
)
def main(file: str, markdown: bool, csv: bool) -> None:
    # MyPy output is in jsonl format (JSON Lines); parse accordingly.
    errors: list[Error] = []
    with open(file) as f:
        for line in f:
            try:
                errors.append(json.loads(line))
            except json.JSONDecodeError:
                pass  # ignore non-JSON lines

    classified_errors = classify_errors(errors)

    summary, total_errors = summarize_errors(classified_errors)

    if csv:
        display_summary_csv(summary, total_errors)
    elif markdown:
        display_summary_markdown(summary, total_errors)
    else:
        display_summary(summary, total_errors)


if __name__ == "__main__":
    main()
